from abc import abstractmethod, abstractclassmethod
import types
from typing import Mapping, Set, Union, Optional, Type, Generator
import functools
import logging
from enum import Enum
import arrow
import datetime
import inspect

from merakicommons.ghost import Ghost, ghost_load_on as _ghost_load_on
from merakicommons.container import SearchableLazyList

from .. import configuration
from ..data import Region, Platform

import json  # Can't use ujson here because of the encoder


LOGGER = logging.getLogger("core")


def ghost_load_on(method):
    return _ghost_load_on(AttributeError)(method)


def add_region_to_kwargs(kwargs):
    region = kwargs.pop("region", None)
    platform = kwargs.pop("platform", None)
    if region is None:
        if platform is None:
            region = configuration.settings.default_region
        else:
            if isinstance(platform, Platform):
                region = platform.region
            else:
                region = Platform(platform).region
    elif not isinstance(region, Region):
        region = Region(region)
    if region is not None:  # region can still be None if the configuration doesn't have a default
        kwargs["region"] = region.value
    return kwargs


def provide_default_region(method):
    @functools.wraps(method)
    def default_region_wrapper(*args, **kwargs):
        kwargs = add_region_to_kwargs(kwargs)
        return method(*args, **kwargs)
    #def default_region_wrapper(self=None, *args, **kwargs):
    #    kwargs = add_region_to_kwargs(kwargs)
    #    if self is not None:
    #        return method(self, *args, **kwargs)
    #    else:
    #        return method(*args, **kwargs)
    return default_region_wrapper


def get_latest_version(region: Union[Region, str], endpoint: Optional[str]):
    from .staticdata.realm import Realms
    if endpoint is not None:
        return Realms(region=region).latest_versions[endpoint]
    else:
        return Realms(region=region).version


class CoreData(object):
    @property
    @abstractclassmethod
    def _renamed(cls) -> Mapping[str, str]:
        pass

    def __init__(self, **kwargs):
        self(**kwargs)

    def __call__(self, **kwargs):
        for key, value in kwargs.items():
            new_key = self._renamed.get(key, key)
            setattr(self, new_key, value)
        return self

    def to_dict(self):
        d = {}
        attrs = {attrname for attrname in dir(self)} - {attrname for attrname in dir(self.__class__)}
        for attr in attrs:
            v = getattr(self, attr)
            if isinstance(v, CoreData):
                v = v.to_dict()
            elif hasattr(v, "__iter__") and not isinstance(v, str):
                if isinstance(v, dict):
                    new_v = {}
                    for k, vi in v.items():
                        if isinstance(vi, CoreData):
                            new_v[k] = vi.to_dict()
                        else:
                            new_v[k] = vi
                    v = new_v
                else:
                    v = [vi.to_dict() if isinstance(vi, CoreData) else vi for vi in v]
            d[attr] = v
        return d


class CoreDataList(list, CoreData):
    def __str__(self):
        return list.__str__(self)

    def __init__(self, *args, **kwargs):
        list.__init__(self, *args)
        CoreData.__init__(self, **kwargs)


class CassiopeiaObject(object):
    _renamed = {}

    def __init__(self, **kwargs):
        # Note: Dto names are not allowed to be passed in.
        self._data = {_type: None for _type in self._data_types}
        # Re-implement __call__ code here so that __call__ can be overridden in subclasses
        results = {_type: {}  for _type in self._data_types}
        found = False
        for key, value in kwargs.items():
            # We don't know which type to put the piece of data under, so put it in any type that supports this key
            for _type in self._data_types:
                if issubclass(_type, CoreData) or key in dir(_type):
                    results[_type][key] = value
                    found = True
            if not found:
                # The user passed in a value that we don't know anything about -- raise a warning.
                LOGGER.warning("When initializing {}, key `{}` is not in type(s) {}. Not set.".format(self.__class__.__name__, key, self._data_types))

        # Now that we've parsed the data and know where to put it all, we can update our data.
        for _type, insert_this in results.items():
            if self._data[_type] is not None:
                self._data[_type] = self._data[_type](**insert_this)
            else:
                self._data[_type] = _type(**insert_this)

    def __str__(self) -> str:
        # This is a bit strange because we'll print a list of dict-like objects rather than one joined dict, but we've decided it's appropriate.
        result = {}
        for _type, data in self._data.items():
            result[str(_type)] = str(data)
        return str(result).replace("\\'", "'")

    @property
    @abstractclassmethod
    def _data_types(cls) -> Set[type]:
        """The `CoreData`_ types that belongs to this core type."""
        pass

    @classmethod
    def from_data(cls, data: CoreData):
        assert data is not None
        self = cls()

        if data.__class__ not in self._data_types:
            raise TypeError("Wrong data type '{}' passed to '{}.from_data'".format(
                data.__class__.__name__, self.__class__.__name__))
        self._data[data.__class__] = data
        return self

    def __call__(self, **kwargs) -> "CassiopeiaObject":
        """Updates `self` with `kwargs` and returns `self`.

        Useful for updating default API parameters, for example:

            for champion in cass.get_champions():
                champion(champData={"tags"}).tags  # only pulls the tag data
        """
        # Update underlying data and deconstruct any Enums the user passed in.
        results = {_type: {}  for _type in self._data_types}
        found = False
        for key, value in kwargs.items():
            # We don't know which type to put the piece of data under, so put it in any type that supports this key
            for _type in self._data_types:
                if issubclass(_type, CoreData) or key in dir(_type):
                    results[_type][key] = value
                    found = True
            if not found:
                # The user passed in a value that we don't know anything about -- raise a warning.
                LOGGER.warning("When initializing {}, key `{}` is not in type(s) {}. Not set.".format(self.__class__.__name__, key, self._data_types))

        # Now that we've parsed the data and know where to put it all, we can update our data.
        for _type, insert_this in results.items():
            if self._data[_type] is not None:
                self._data[_type] = self._data[_type](**insert_this)
            else:
                self._data[_type] = _type(**insert_this)
        return self

    def to_dict(self):
        d = {}
        for data_type in self._data_types:
            new = self._data[data_type].to_dict()
            d.update(new)
        return d

    def to_json(self, **kwargs):
        return json.dumps(self.to_dict(), cls=CassiopeiaJsonEncoder, **kwargs)

    def __json__(self, **kwargs):
        return self.to_json(**kwargs)


class GetFromPipeline(type):
    def __call__(cls: "CassiopeiaPipelineObject", *args, **kwargs):
        if 'region' in inspect.signature(cls.__get_query_from_kwargs__).parameters:
            kwargs = add_region_to_kwargs(kwargs)
        pipeline = configuration.settings.pipeline
        query = cls.__get_query_from_kwargs__(**kwargs)
        if hasattr(cls, "version") and query.get("version", None) is None and cls.__name__ not in ["Realms", "Match"]:
            query["version"] = get_latest_version(region=query["region"], endpoint=None)
        return pipeline.get(cls, query=query)


class CassiopeiaPipelineObject(CassiopeiaObject, metaclass=GetFromPipeline):
    @classmethod
    def _construct_normally(cls, *args, **kwargs) -> "CassiopeiaObject":
        # cls.__class__ will be this class's metaclass (GetFromPipeline), so supering that will find the metaclass's
        # class, which is `type` in this case. Then `type`'s `__call__` will be called.
        # This has the effect of skipping the GetFromPipeline instantiation, and defaulting to the normal class
        # instantiation for the class.
        # TODO I don't know how to handle includedData -> included_data for core...
        if "includedData" in kwargs:
            kwargs["included_data"] = kwargs.pop("includedData")
        return super(cls.__class__, cls).__call__(*args, **kwargs)

    @abstractmethod
    def __get_query__(self):
        pass

    @classmethod
    @provide_default_region
    def __get_query_from_kwargs__(cls, **kwargs):
        return kwargs

    def __hash__(self):
        stuff = sorted(self.__get_query__().items())
        for i, (key, value) in enumerate(stuff):
            if isinstance(value, set):
                stuff[i] = (key, tuple(value))
        return hash(tuple(stuff))

    def __eq__(self, other: "CassiopeiaPipelineObject") -> bool:
        if not isinstance(other, self.__class__):  # Using isinstance here seems bad
            return False
        else:
            return hash(self) == hash(other)

    def __ne__(self, other: "CassiopeiaPipelineObject") -> bool:
        return not self == other


class CassiopeiaGhost(CassiopeiaPipelineObject, Ghost):
    def load(self, load_groups: Set = None) -> "CassiopeiaGhost":
        if load_groups is None:
            load_groups = self._Ghost__load_groups
        if self._Ghost__all_loaded:
            return self
        self.__load__()
        for load_group in load_groups:
            self._Ghost__set_loaded(load_group)  # __load__ doesn't trigger __set_loaded.
        return self

    def __load__(self, load_group: CoreData = None, load_groups: Set = None) -> None:
        if load_groups is None:
            load_groups = self._Ghost__load_groups
        if load_group is None:  # Load all groups
            if self._Ghost__all_loaded:
                raise ValueError("object has already been loaded.")
            for group in load_groups:
                if not self._Ghost__is_loaded(group):
                    self.__load__(group)
        else:  # Load the specific load group
            if self._Ghost__is_loaded(load_group):
                raise ValueError("object has already been loaded.")
            query = self.__get_query__()
            if hasattr(self.__class__, "version") and "version" not in query and not self.__class__.__name__ == "Realms":
                query["version"] = get_latest_version(region=query["region"], endpoint=None)
            data = configuration.settings.pipeline.get(type=self._load_types[load_group], query=query)
            self.__load_hook__(load_group, data)

    @property
    def _load_types(cls):
        return {t: t for t in cls._data_types}

    @property
    def _load_type(cls):
        return cls

    @classmethod
    def from_data(cls, data: CoreData, loaded_groups: Optional[Set[Type[CoreData]]] = None):
        assert data is not None
        # Manually skip the CheckCache (well, all metaclass' __call__s) for ghost objects if they are
        # created via this constructor.
        self = cls._construct_normally()

        # Make spots for the data and put it in
        self._data = {_type: None for _type in self._data_types}
        if data.__class__ not in self._data_types:
            raise TypeError("Wrong data type '{}' passed to '{}.from_data'".format(
                data.__class__.__name__, self.__class__.__name__))
        self._data[data.__class__] = data

        # Set as loaded
        if loaded_groups is None:
            for load_group in self._Ghost__load_groups:
                self._Ghost__set_loaded(load_group)
        else:
            for load_group in loaded_groups:
                self._Ghost__set_loaded(load_group)
        return self

    def __load_hook__(self, load_group: CoreData, data: CoreData) -> None:
        if not isinstance(data, CoreData):
            raise TypeError("expected subclass of CoreData, got {cls}".format(cls=data.__class__))
        self._data[load_group] = data


class CassiopeiaLazyList(SearchableLazyList, CassiopeiaPipelineObject):
    def __init__(self, *args, **kwargs):
        if "generator" in kwargs:
            generator = kwargs.pop("generator")
        else:
            if len(args) == 1 and isinstance(args[0], types.GeneratorType):
                generator = args[0]
            else:
                def generator(*args):
                    for arg in args:
                        yield arg
                generator = generator(args)
        SearchableLazyList.__init__(self, generator)
        # Something feels very wrong; this is meant to work with MatchHistory.from_generator
        if self.__class__ is not CassiopeiaLazyList:
            self.__init__(**kwargs)
        else:
            CassiopeiaObject.__init__(self, **kwargs)

    @classmethod
    def from_data(cls, *args, **kwargs):
        return cls._construct_normally(*args, **kwargs)

    @classmethod
    def from_generator(cls, generator: Generator, **kwargs):
        self = cls.__new__(cls)
        CassiopeiaLazyList.__init__(self, generator=generator, **kwargs)
        return self

    def __hash__(self):
        return id(self)

    def __str__(self):
        return SearchableLazyList.__str__(self)


class CassiopeiaJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Enum):
            return obj.name
        elif isinstance(obj, (datetime.datetime, arrow.Arrow)):
            return obj.isoformat()
        elif isinstance(obj, datetime.timedelta):
            return obj.seconds
        return json.JSONEncoder.default(self, obj)



from typing import List, Set, Dict, Union, TextIO
import arrow
import datetime

from .data import Region, Queue, Season, Tier, Division, Position
from .core import Champion, Summoner, ChampionMastery, Rune, Item, Match, Map, SummonerSpell, Realms, ProfileIcon, LanguageStrings, CurrentMatch, ShardStatus, Versions, MatchHistory, Champions, ChampionMasteries, Runes, Items, SummonerSpells, Maps, FeaturedMatches, Locales, ProfileIcons, ChallengerLeague, GrandmasterLeague, MasterLeague, League, LeagueSummonerEntries, LeagueEntries, Patch, VerificationString, ChampionRotation
from .datastores import common as _common_datastore
from ._configuration import Settings, load_config, get_default_config
from . import configuration


# Settings endpoints

def apply_settings(config: Union[str, TextIO, Dict, Settings]):
    if not isinstance(config, (Dict, Settings)):
        config = load_config(config)
    if not isinstance(config, Settings):
        settings = Settings(config)
    else:
        settings = config

    # Load any plugins after everything else has finished importing
    import importlib
    for plugin in settings.plugins:
        imported_plugin = importlib.import_module("cassiopeia.plugins.{plugin}.monkeypatch".format(plugin=plugin))

    print_calls(settings._Settings__default_print_calls, settings._Settings__default_print_riot_api_key)

    # Overwrite the old settings
    configuration._settings = settings

    # Initialize the pipeline immediately
    _ = configuration.settings.pipeline


def set_riot_api_key(key: str):
    configuration.settings.set_riot_api_key(key)


def set_default_region(region: Union[Region, str]):
    configuration.settings.set_region(region)


def print_calls(calls: bool, api_key: bool = False):
    _common_datastore._print_calls = calls
    _common_datastore._print_api_key = api_key


# Data endpoints

def get_league_entries(summoner: Summoner) -> LeagueEntries:
    return summoner.league_entries


def get_paginated_league_entries(queue: Queue, tier: Tier, division: Division, region: Union[Region, str] = None) -> LeagueEntries:
    return LeagueEntries(region=region, queue=queue, tier=tier, division=division)


def get_master_league(queue: Union[Queue, int, str], region: Union[Region, str] = None) -> MasterLeague:
    return MasterLeague(queue=queue, region=region)

def get_grandmaster_league(queue: Union[Queue, int, str], region: Union[Region, str] = None) -> GrandmasterLeague:
    return GrandmasterLeague(queue=queue, region=region)

def get_challenger_league(queue: Union[Queue, int, str], region: Union[Region, str] = None) -> ChallengerLeague:
    return ChallengerLeague(queue=queue, region=region)


def get_match_history(summoner: Summoner, begin_index: int = None, end_index: int = None, begin_time: arrow.Arrow = None, end_time: arrow.Arrow = None, queues: Set[Queue] = None, seasons: Set[Season] = None, champions: Set[Champion] = None):
    return MatchHistory(summoner=summoner, begin_index=begin_index, end_index=end_index, begin_time=begin_time, end_time=end_time, queues=queues, seasons=seasons, champions=champions)

def get_match(id, region: Union[Region, str] = None) -> Match:
    return Match(id=id, region=region)


def get_featured_matches(region: Union[Region, str] = None) -> FeaturedMatches:
    return FeaturedMatches(region=region)


def get_current_match(summoner: Summoner, region: Union[Region, str] = None) ->  CurrentMatch:
    return CurrentMatch(summoner=summoner, region=region)


def get_champion_masteries(summoner: Summoner, region: Union[Region, str] = None) -> ChampionMasteries:
    return ChampionMasteries(summoner=summoner, region=region)


def get_champion_mastery(summoner: Summoner, champion: Union[Champion, int, str], region: Union[Region, str] = None) -> ChampionMastery:
    return ChampionMastery(champion=champion, summoner=summoner, region=region)


def get_summoner(*, id: str = None, account_id: str = None, name: str = None, region: Union[Region, str] = None) -> Summoner:
    return Summoner(id=id, account_id=account_id, name=name, region=region)


def get_champion(key: Union[str, int], region: Union[Region, str] = None) -> Champion:
    return get_champions(region=region)[key]


def get_champions(region: Union[Region, str] = None) -> Champions:
    return Champions(region=region)


def get_runes(region: Union[Region, str] = None) -> Runes:
    return Runes(region=region)


def get_summoner_spells(region: Union[Region, str] = None) -> SummonerSpells:
    return SummonerSpells(region=region)


def get_items(region: Union[Region, str] = None) -> Items:
    return Items(region=region)


def get_maps(region: Union[Region, str] = None) -> Maps:
    return Maps(region=region)


def get_profile_icons(region: Union[Region, str] = None) -> ProfileIcons:
    return ProfileIcons(region=region)


def get_realms(region: Union[Region, str] = None) -> Realms:
    return Realms(region=region)


def get_status(region: Union[Region, str] = None) -> ShardStatus:
    return ShardStatus(region=region)


def get_language_strings(region: Union[Region, str] = None) -> LanguageStrings:
    return LanguageStrings(region=region)


def get_locales(region: Union[Region, str] = None) -> List[str]:
    return Locales(region=region)


def get_versions(region: Union[Region, str] = None) -> List[str]:
    return Versions(region=region)


def get_version(date: datetime.date = None, region: Union[Region, str] = None) -> Union[None, str]:
    versions = get_versions(region)
    if date is None:
        return versions[0]
    else:
        patch = Patch.from_date(date, region=region)
        for version in versions:
            if patch.majorminor in version:
                return version
    return None


def get_verification_string(summoner: Summoner) -> VerificationString:
    return VerificationString(summoner=summoner)


def get_champion_rotations(region: Union[Region, str] = None) -> ChampionRotation:
    return ChampionRotation(region=region)


# Pipeline

def _get_pipeline():
    return configuration.settings.pipeline



from typing import List, Union

from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable

from ...data import Region, Platform
from ..common import CoreData, CassiopeiaObject, CassiopeiaGhost, CoreDataList, CassiopeiaLazyList, get_latest_version, provide_default_region, ghost_load_on
from .common import ImageData, Sprite, Image
from ...dto.staticdata import map as dto


##############
# Data Types #
##############


class MapListData(CoreDataList):
    _dto_type = dto.MapListDto
    _renamed = {"included_data": "includedData"}


class MapData(CoreData):
    _dto_type = dto.MapDto
    _renamed = {"mapId": "id", "mapName": "name", "unpurchasableItemList": "unpurchasableItems", "included_data": "includedData"}

    def __call__(self, **kwargs):
        if "image" in kwargs:
            self.image = ImageData(**kwargs.pop("image"))
        super().__call__(**kwargs)
        return self


##############
# Core Types #
##############


class Maps(CassiopeiaLazyList):
    _data_types = {MapListData}

    @provide_default_region
    def __init__(self, *, region: Union[Region, str] = None, version: str = None, locale: str = None):
        if locale is None and region is not None:
            locale = Region(region).default_locale
        kwargs = {"region": region, "locale": locale}
        if version is not None:
            kwargs["version"] = version
        CassiopeiaObject.__init__(self, **kwargs)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[MapListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        try:
            return self._data[MapListData].version
        except AttributeError:
            version = get_latest_version(region=self.region, endpoint="map")
            self(version=version)
            return self._data[MapListData].version

    @property
    def locale(self) -> str:
        return self._data[MapListData].locale


@searchable({str: ["name", "locale"], int: ["id"]})
class Map(CassiopeiaGhost):
    _data_types = {MapData}

    @provide_default_region
    def __init__(self, *, id: int = None, name: str = None, region: Union[Region, str] = None, version: str = None, locale: str = None):
        if locale is None and region is not None:
            locale = Region(region).default_locale
        kwargs = {"region": region, "locale": locale}
        if version is not None:
            kwargs["version"] = version
        if id is not None:
            kwargs["id"] = id
        if name is not None:
            kwargs["name"] = name
        super().__init__(**kwargs)

    def __get_query__(self):
        query = {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale}
        if hasattr(self._data[MapData], "id"):
            query["id"] = self.id
        if hasattr(self._data[MapData], "name"):
            query["name"] = self.name
        return query

    def __eq__(self, other: "Map"):
        if not isinstance(other, Map) or self.region != other.region:
            return False
        s = {}
        o = {}
        if hasattr(self._data[MapData], "id"): s["id"] = self.id
        if hasattr(other._data[MapData], "id"): o["id"] = other.id
        if hasattr(self._data[MapData], "name"): s["name"] = self.name
        if hasattr(other._data[MapData], "name"): o["name"] = other.name
        if any(s.get(key, "s") == o.get(key, "o") for key in s):
            return True
        else:
            return self.id == other.id

    def __str__(self):
        region = self.region
        id_ = "?"
        name = "?"
        if hasattr(self._data[MapData], "id"):
            id_ = self.id
        if hasattr(self._data[MapData], "name"):
            name = self.name
        return "Map(name='{name}', id={id_}, region='{region}')".format(name=name, id_=id_, region=region.value)

    __hash__ = CassiopeiaGhost.__hash__

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[MapData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this map."""
        try:
            return self._data[MapData].version
        except AttributeError:
            version = get_latest_version(region=self.region, endpoint="map")
            self(version=version)
            return self._data[MapData].version

    @property
    def locale(self) -> str:
        """The locale for this map."""
        return self._data[MapData].locale or self.region.default_locale

    @CassiopeiaGhost.property(MapData)
    @ghost_load_on
    def id(self) -> int:
        """The map's ID."""
        return self._data[MapData].id

    @CassiopeiaGhost.property(MapData)
    @ghost_load_on
    def name(self) -> str:
        return self._data[MapData].name

    @CassiopeiaGhost.property(MapData)
    @ghost_load_on
    def unpurchasable_items(self) -> List[int]:
        return self._data[MapData].unpurchasableItems

    @CassiopeiaGhost.property(MapData)
    @ghost_load_on
    @lazy
    def image(self) -> Image:
        image = Image.from_data(self._data[MapData].image)
        image(version=self.version)
        return image

    @lazy_property
    def sprite(self) -> Sprite:
        return self.image.spriteInfo



from .settings import Settings, get_default_config
from .load import load_config


class MetaConfiguration(type):
    # A singleton-like metaclass that never changes the object, but instead modifies the underlying settings attribute
    # upon instantiation.
    # This is required in order to not lose the reference that gets defined in cassiopeia/__init__.py.
    def __init__(self, *args, **kwargs):
        self.__instance = None
        super().__init__(*args, **kwargs)

    def __call__(self, settings: Settings = None):
        if self.__instance is None:
            self.__instance = super().__call__(settings)
            return self.__instance
        else:
            self.__instance._settings = settings
            return self.__instance


class CassiopeiaConfiguration(object, metaclass=MetaConfiguration):
    def __init__(self, settings: Settings = None):
        self._settings = settings

    @property
    def settings(self):
        if self._settings is None:
            config = load_config()  # Use default
            settings = Settings(config)
            self._settings = settings
        return self._settings


import sys
sys.modules["cassiopeia.configuration"] = CassiopeiaConfiguration()



from .common import DtoObject


class MiniSeriesDto(DtoObject):
    pass


class LeagueEntryDto(DtoObject):
    pass


class LeagueDto(DtoObject):
    pass


class LeagueSummonerEntriesDto(DtoObject):
    pass


class LeagueEntriesDto(DtoObject):
    pass


class ChallengerLeagueListDto(DtoObject):
    pass


class GrandmasterLeagueListDto(DtoObject):
    pass


class MasterLeagueListDto(DtoObject):
    pass



import re
import zlib
from contextlib import contextmanager, ExitStack
from io import BytesIO
from typing import Mapping, MutableMapping, Any, Union, Dict, List
from urllib.parse import urlencode

try:
    from pycurl import Curl
    USE_PYCURL = True
except ImportError:
    import requests
    USE_PYCURL = False
    Curl = None  # This might break a few type hints but they are all internal and not user-facing.

from merakicommons.ratelimits import RateLimiter

try:
    import certifi
except ImportError:
    certifi = None

try:
    import ujson as json
except ImportError:
    import json


_print_calls = True
_print_api_key = False


class HTTPError(RuntimeError):
    def __init__(self, message, code, response_headers: Dict[str, str] = None):
        super().__init__(message)
        self.code = code
        self.response_headers = response_headers or {}


if USE_PYCURL:
    class HTTPClient(object):
        @staticmethod
        def _execute(curl: Curl, close_connection: bool) -> int:
            curl.perform()
            status_code = curl.getinfo(curl.HTTP_CODE)
            if close_connection:
                curl.close()
            return status_code

        @staticmethod
        def _get(url: str, headers: Mapping[str, str] = None, rate_limiters: List[RateLimiter] = None, connection: Curl = None) -> (int, bytes, dict):
            if not headers:
                request_headers = ["Accept-Encoding: gzip"]
            else:
                request_headers = ["{header}: {value}".format(header=key, value=value) for key, value in headers.items()]
                if "Accept-Encoding" not in headers:
                    request_headers.append("Accept-Encoding: gzip")

            response_headers = {}

            def get_response_headers(header_line: bytes) -> None:
                header_line = header_line.decode("ISO-8859-1")

                if ":" not in header_line:
                    return

                name, value = header_line.split(":", 1)
                response_headers[name.strip()] = value.strip()

            buffer = BytesIO()

            curl = connection if connection is not None else Curl()

            curl.setopt(curl.URL, url)
            curl.setopt(curl.WRITEDATA, buffer)
            curl.setopt(curl.HEADERFUNCTION, get_response_headers)
            curl.setopt(curl.HTTPHEADER, request_headers)
            if certifi:
                curl.setopt(curl.CAINFO, certifi.where())

            if _print_calls:
                _url = url
                if isinstance(_url, bytes):
                    _url = str(_url)[2:-1]
                if _print_api_key and ".api.riotgames.com/lol" in _url:
                    if "?" not in _url:
                        _url += "?api_key={}".format(headers["X-Riot-Token"])
                    else:
                        _url += "&api_key={}".format(headers["X-Riot-Token"])
                print("Making call: {}".format(_url))
            if rate_limiters:
                with ExitStack() as stack:
                    # Enter each context manager / rate limiter
                    limiters = [stack.enter_context(rate_limiter) for rate_limiter in rate_limiters]
                    exit_limiters = stack.pop_all().__exit__
                    status_code = HTTPClient._execute(curl, connection is None)
                exit_limiters(None, None, None)
            else:
                status_code = HTTPClient._execute(curl, connection is None)

            body = buffer.getvalue()

            # Decompress if we got gzipped data
            try:
                content_encoding = response_headers["Content-Encoding"].upper()
                if "GZIP" == content_encoding:
                    body = zlib.decompress(body, zlib.MAX_WBITS | 16)
            except KeyError:
                pass

            return status_code, body, response_headers

        def get(self, url: str, parameters: MutableMapping[str, Any] = None, headers: Mapping[str, str] = None, rate_limiters: List[RateLimiter] = None, connection: Curl = None, encode_parameters: bool = True) -> (Union[dict, list, str, bytes], dict):
            if parameters:
                if encode_parameters:
                    parameters = {k: str(v).lower() if isinstance(v, bool) else v for k, v in parameters.items()}
                    parameters = urlencode(parameters, doseq=True)
                url = "{url}?{params}".format(url=url, params=parameters)

            status_code, body, response_headers = HTTPClient._get(url, headers, rate_limiters, connection)

            content_type = response_headers.get("Content-Type", "application/octet-stream").upper()

            # Decode to text if a charset is included
            match = re.search("CHARSET=(\S+)", content_type)
            if match:
                encoding = match.group(1)
                body = body.decode(encoding)

                # Load JSON if necessary
                if "APPLICATION/JSON" in content_type:
                    body = json.loads(body)

            # Handle errors
            if status_code >= 400:
                if isinstance(body, dict):
                    message = body.get("status", {}).get("message", "")
                elif isinstance(body, str):
                    message = body
                else:
                    message = ""

                raise HTTPError(message, status_code, response_headers)

            return body, response_headers

        @contextmanager
        def new_session(self) -> Curl:
            session = Curl()
            yield session
            session.close()

else:  # Use requests
    class HTTPClient(object):
        @staticmethod
        def _get(url: str, headers: Mapping[str, str] = None, rate_limiters: List[RateLimiter] = None) -> (int, bytes, dict):
            if not headers:
                request_headers = {"Accept-Encoding": "gzip"}
            else:
                request_headers = {k: v for k, v in headers.items()}
                if "Accept-Encoding" not in headers:
                    request_headers["Accept-Encoding"] = "gzip"

            if _print_calls:
                _url = url
                if isinstance(_url, bytes):
                    _url = str(_url)[2:-1]
                if _print_api_key and ".api.riotgames.com/lol" in _url:
                    if "?" not in _url:
                        _url += "?api_key={}".format(headers["X-Riot-Token"])
                    else:
                        _url += "&api_key={}".format(headers["X-Riot-Token"])
                print("Making call: {}".format(_url))
            if rate_limiters:
                with ExitStack() as stack:
                    # Enter each context manager / rate limiter
                    limiters = [stack.enter_context(rate_limiter) for rate_limiter in rate_limiters]
                    exit_limiters = stack.pop_all().__exit__
                    r = requests.get(url, headers=request_headers)
                exit_limiters(None, None, None)
            else:
                r = requests.get(url, headers=request_headers)

            return r

        def get(self, url: str, parameters: MutableMapping[str, Any] = None, headers: Mapping[str, str] = None, rate_limiters: List[RateLimiter] = None, connection: Curl = None, encode_parameters: bool = True) -> (Union[dict, list, str, bytes], dict):
            if parameters:
                if encode_parameters:
                    parameters = {k: str(v).lower() if isinstance(v, bool) else v for k, v in parameters.items()}
                    parameters = urlencode(parameters, doseq=True)
                url = "{url}?{params}".format(url=url, params=parameters)

            #status_code, body, response_headers = HTTPClient._get(url, headers, rate_limiters)
            r = HTTPClient._get(url, headers, rate_limiters)
            response_headers = r.headers

            # Handle errors
            if r.status_code >= 400:
                raise HTTPError(r.reason, r.status_code, response_headers)

            content_type = response_headers.get("Content-Type", "application/octet-stream").upper()

            body = r.text
            # Decode to text if a charset is included
            match = re.search("CHARSET=(\S+)", content_type)
            if match:
                # Load JSON if necessary
                if "APPLICATION/JSON" in content_type:
                    body = r.json()
            if "IMAGE/" in content_type:
                body = r.content

            return body, response_headers

        @contextmanager
        def new_session(self) -> requests.Session:
            session = requests.Session()
            yield session
            session.close()



from typing import List, Set, Union
from PIL.Image import Image as PILImage

from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable

from ... import configuration
from ...data import Region, Platform
from ..common import CoreData, CoreDataList, CassiopeiaObject, CassiopeiaGhost, CassiopeiaLazyList, get_latest_version, provide_default_region, ghost_load_on
from .common import Image
from ...dto.staticdata import rune as dto


##############
# Data Types #
##############


class RunePathData(CoreData):
    _renamed = {}


class RuneListData(CoreDataList):
    _dto_type = dto.RuneListDto
    _renamed = {"included_data": "includedData"}


class RuneData(CoreData):
    _dto_type = dto.RuneDto
    _renamed = {"longDesc": "longDescription", "shortDesc": "shortDescription", "included_data": "includedData"}

    def __call__(self, **kwargs):
        if "icon" in kwargs:
            self.image = RuneImageData(icon=kwargs.pop("icon"))
        super().__call__(**kwargs)
        return self


class RuneImageData(CoreData):
    _renamed = {}


##############
# Core Types #
##############


class RunePath(CassiopeiaObject):
    _data_types = {RunePathData}

    def __init__(self, *, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if included_data is None:
            included_data = {"all"}
        if locale is None and region is not None:
            locale = Region(region).default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if version:
            kwargs["version"] = version
        CassiopeiaObject.__init__(self, **kwargs)

    @property
    def id(self):
        return self._data[RunePathData].id

    @property
    def name(self):
        return self._data[RunePathData].name

    @property
    def key(self):
        return self._data[RunePathData].key

    @property
    def image_url(self):
        url = "https://ddragon.leagueoflegends.com/cdn/img/perk-images/" + self._data[RunePathData].icon
        return url


class Runes(CassiopeiaLazyList):
    _data_types = {RuneListData}

    @provide_default_region
    def __init__(self, *, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if included_data is None:
            included_data = {"all"}
        if locale is None and region is not None:
            locale = Region(region).default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if version:
            kwargs["version"] = version
        CassiopeiaObject.__init__(self, **kwargs)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[RuneListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        try:
            return self._data[RuneListData].version
        except AttributeError:
            version = get_latest_version(region=self.region, endpoint="rune")
            self(version=version)
            return self._data[RuneListData].version

    @property
    def locale(self) -> str:
        """The locale for this champion."""
        return self._data[RuneListData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additonal information for this champion when it's loaded."""
        return self._data[RuneListData].includedData

    @property
    def precision(self) -> List["Rune"]:
        return self.filter(lambda rune: rune.path.key == "Precision")

    @property
    def domination(self) -> List["Rune"]:
        return self.filter(lambda rune: rune.path.key == "Domination")

    @property
    def sorcery(self) -> List["Rune"]:
        return self.filter(lambda rune: rune.path.key == "Sorcery")

    @property
    def resolve(self) -> List["Rune"]:
        return self.filter(lambda rune: rune.path.key == "Resolve")

    @property
    def inspiration(self) -> List["Rune"]:
        return self.filter(lambda rune: rune.path.key == "Inspiration")

    @property
    def keystones(self) -> List["Rune"]:
        return self.filter(lambda rune: rune.is_keystone)


class RuneImage(CassiopeiaObject):
    _data_types = {RuneImageData}

    @property
    def url(self) -> str:
        return "https://ddragon.leagueoflegends.com/cdn/img/{icon}".format(icon=self._data[RuneImageData].icon)

    @lazy_property
    def image(self) -> PILImage:
        return configuration.settings.pipeline.get(PILImage, query={"url": self.url})


@searchable({str: ["name", "tags", "path", "region", "platform", "locale"], int: ["id"], RunePath: ["path"], Region: ["region"], Platform: ["platform"]})
class Rune(CassiopeiaGhost):
    _data_types = {RuneData}

    @provide_default_region
    def __init__(self, *, id: int = None, name: str = None, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if included_data is None:
            included_data = {"all"}
        if locale is None and region is not None:
            locale = Region(region).default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if id is not None:
            kwargs["id"] = id
        if name is not None:
            kwargs["name"] = name
        if version is not None:
            kwargs["version"] = version
        super().__init__(**kwargs)

    def __get_query__(self):
        query = {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}
        if hasattr(self._data[RuneData], "id"):
            query["id"] = self._data[RuneData].id
        if hasattr(self._data[RuneData], "name"):
            query["name"] = self._data[RuneData].name
        return query

    def __eq__(self, other: "Rune"):
        if not isinstance(other, Rune) or self.region != other.region:
            return False
        s = {}
        o = {}
        if hasattr(self._data[RuneData], "id"): s["id"] = self.id
        if hasattr(other._data[RuneData], "id"): o["id"] = other.id
        if hasattr(self._data[RuneData], "name"): s["name"] = self.name
        if hasattr(other._data[RuneData], "name"): o["name"] = other.name
        if any(s.get(key, "s") == o.get(key, "o") for key in s):
            return True
        else:
            return self.id == other.id

    def __str__(self):
        region = self.region
        id_ = "?"
        name = "?"
        if hasattr(self._data[RuneData], "id"):
            id_ = self.id
        if hasattr(self._data[RuneData], "name"):
            name = self.name
        return "Rune(name='{name}', id={id_}, region='{region}')".format(name=name, id_=id_, region=region.value)

    __hash__ = CassiopeiaGhost.__hash__

    @lazy_property
    def region(self) -> Region:
        """The region for this rune."""
        return Region(self._data[RuneData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this rune."""
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this rune."""
        try:
            return self._data[RuneData].version
        except AttributeError:
            version = get_latest_version(region=self.region, endpoint="rune")
            self(version=version)
            return self._data[RuneData].version

    @property
    def locale(self) -> str:
        """The locale for this rune."""
        return self._data[RuneData].locale or self.region.default_locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additional information for this champion when it's loaded."""
        return self._data[RuneData].includedData

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on
    def path(self) -> RunePath:
        data = RunePathData(**self._data[RuneData].path)  # This seems out of place but we never request a RunePath so it never geos through the pipeline
        return RunePath.from_data(data=data)

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on
    def tier(self) -> int:
        return self._data[RuneData].tier

    @property
    def is_keystone(self) -> bool:
        return self.tier == 0

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on
    def name(self) -> str:
        """The rune's name."""
        return self._data[RuneData].name

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on
    def id(self) -> int:
        """The rune's ID."""
        return self._data[RuneData].id

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on
    def short_description(self) -> str:
        return self._data[RuneData].shortDescription

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on
    def long_description(self) -> str:
        return self._data[RuneData].longDescription

    @CassiopeiaGhost.property(RuneData)
    @ghost_load_on
    @lazy
    def image(self) -> Image:
        """The image information for this rune."""
        image = RuneImage.from_data(self._data[RuneData].image)
        image(version=self.version)
        return image



from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.league import LeagueEntry, LeagueEntries, LeagueEntriesData, LeagueSummonerEntries, LeagueSummonerEntriesData, League, ChallengerLeagueListData, ChallengerLeague, GrandmasterLeagueListData, GrandmasterLeague, MasterLeagueListData, MasterLeague, LeagueData, LeagueEntryData

from ..dto.league import LeagueDto, LeagueEntriesDto, LeagueSummonerEntriesDto, ChallengerLeagueListDto, GrandmasterLeagueListDto, MasterLeagueListDto, LeagueEntryDto

T = TypeVar("T")
F = TypeVar("F")


class LeagueTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(LeagueDto, LeagueData)
    def league_dto_to_data(self, value: LeagueDto, context: PipelineContext = None) -> LeagueData:
        return LeagueData(**value)

    @transform.register(LeagueEntryDto, LeagueEntryData)
    def league_entry_dto_to_data(self, value: LeagueEntryDto, context: PipelineContext = None) -> LeagueEntryData:
        return LeagueEntryData(**value)

    @transform.register(LeagueSummonerEntriesDto, LeagueSummonerEntriesData)
    def leagues_summoner_entries_dto_to_data(self, value: LeagueSummonerEntriesDto, context: PipelineContext = None) -> LeagueSummonerEntriesData:
        data = deepcopy(value)
        for entry in data["entries"]:
            entry["region"] = data["region"]
        data = [LeagueTransformer.league_entry_dto_to_data(self, entry) for entry in data["entries"]]
        return LeagueSummonerEntriesData(data, summoner_id=value["summonerId"], region=value["region"])

    @transform.register(LeagueEntriesDto, LeagueEntriesData)
    def leagues_entries_dto_to_data(self, value: LeagueEntriesDto, context: PipelineContext = None) -> LeagueEntriesData:
        kwargs = {
            "region": value["region"],
            "queue": value["queue"],
            "tier": value["tier"],
            "division": value["division"],
            "page": value["page"]
        }
        return LeagueEntriesData([self.league_entry_dto_to_data(entry) for entry in value["entries"]], **kwargs)

    @transform.register(ChallengerLeagueListDto, ChallengerLeagueListData)
    def challenger_league_list_dto_to_data(self, value: ChallengerLeagueListDto, context: PipelineContext = None) -> ChallengerLeagueListData:
        return ChallengerLeagueListData(**value)

    @transform.register(GrandmasterLeagueListDto, GrandmasterLeagueListData)
    def grandmaster_league_list_dto_to_data(self, value: GrandmasterLeagueListDto, context: PipelineContext = None) -> GrandmasterLeagueListData:
        return GrandmasterLeagueListData(**value)

    @transform.register(MasterLeagueListDto, MasterLeagueListData)
    def master_league_list_dto_to_data(self, value: MasterLeagueListDto, context: PipelineContext = None) -> MasterLeagueListData:
        return MasterLeagueListData(**value)

    # Data to Core

    def league_data_to_core(self, value: LeagueData, context: PipelineContext = None) -> League:
        data = deepcopy(value)
        return League.from_data(data)

    @transform.register(LeagueEntryData, LeagueEntry)
    def league_entry_data_to_core(self, value: LeagueEntryData, context: PipelineContext = None) -> LeagueEntry:
        data = deepcopy(value)
        return LeagueEntry.from_data(data=data)

    @transform.register(LeagueEntriesData, LeagueEntries)
    def league_entries_data_to_core(self, value: LeagueEntriesData, context: PipelineContext = None) -> LeagueEntries:
        return LeagueEntries(*[LeagueTransformer.league_entry_data_to_core(self, entry) for entry in value], tier=value.tier, division=value.division, queue=value.queue, region=value.region)

    #@transform.register(ChallengerLeagueListData, ChallengerLeague)
    def challenger_league_list_data_to_core(self, value: ChallengerLeagueListData, context: PipelineContext = None) -> ChallengerLeague:
        data = deepcopy(value)
        return ChallengerLeague.from_data(data)

    #@transform.register(GrandmasterLeagueListData, GrandmasterLeague)
    def grandmaster_league_list_data_to_core(self, value: GrandmasterLeagueListData, context: PipelineContext = None) -> GrandmasterLeague:
        data = deepcopy(value)
        return GrandmasterLeague.from_data(data)

    #@transform.register(MasterLeagueListData, MasterLeague)
    def master_league_list_data_to_core(self, value: MasterLeagueListData, context: PipelineContext = None) -> MasterLeague:
        data = deepcopy(value)
        return MasterLeague.from_data(data)



from datapipelines import CompositeDataTransformer

from .staticdata import StaticDataTransformer
from .champion import ChampionTransformer
from .championmastery import ChampionMasteryTransformer
from .summoner import SummonerTransformer
from .match import MatchTransformer
from .spectator import SpectatorTransformer
from .status import StatusTransformer
from .leagues import LeagueTransformer
from .thirdpartycode import ThirdPartyCodeTransformer


riotapi_transformer = CompositeDataTransformer([
    StaticDataTransformer(),
    ChampionTransformer(),
    ChampionMasteryTransformer(),
    SummonerTransformer(),
    MatchTransformer(),
    SpectatorTransformer(),
    StatusTransformer(),
    LeagueTransformer(),
    ThirdPartyCodeTransformer()
])

__transformers__ = [riotapi_transformer]



from enum import Enum
import arrow


class Region(Enum):
    brazil = "BR"
    europe_north_east = "EUNE"
    europe_west = "EUW"
    japan = "JP"
    korea = "KR"
    latin_america_north = "LAN"
    latin_america_south = "LAS"
    north_america = "NA"
    oceania = "OCE"
    turkey = "TR"
    russia = "RU"

    @property
    def platform(self) -> "Platform":
        return getattr(Platform, self.name)

    @property
    def default_locale(self) -> str:
        return DEFAULT_LOCALE[self]

    @staticmethod
    def from_platform(platform):
        try:
            return platform.region
        except AttributeError:
            return Platform(platform).region

    @property
    def timezone(self) -> str:
        tzs = {
            'NA': 'GMT-8',
            'LAN': 'GMT-7',
            'LAS': 'GMT-5',
            'BR': 'GMT-4',
            'EUW': 'GMT-2',
            'TR': 'GMT-0',
            'EUNE': 'GMT+1',
            'RU': 'GMT+3',
            'KR': 'GMT+6',
            'JP': 'GMT+7',
            'OCE': 'GMT+8',
        }
        return tzs[self.value]


class Platform(Enum):
    brazil = "BR1"
    europe_north_east = "EUN1"
    europe_west = "EUW1"
    japan = "JP1"
    korea = "KR"
    latin_america_north = "LA1"
    latin_america_south = "LA2"
    north_america = "NA1"
    oceania = "OC1"
    turkey = "TR1"
    russia = "RU"

    @property
    def region(self) -> "Region":
        return getattr(Region, self.name)

    @property
    def default_locale(self) -> str:
        return DEFAULT_LOCALE[self]

    @staticmethod
    def from_region(region):
        try:
            return region.platform
        except AttributeError:
            return Region(region).platform


DEFAULT_LOCALE = {
    Region.brazil: "pt_BR",
    Platform.brazil: "pt_BR",
    Region.europe_north_east: "en_GB",
    Platform.europe_north_east: "en_GB",
    Region.europe_west: "en_GB",
    Platform.europe_west: "en_GB",
    Region.japan: "ja_JP",
    Platform.japan: "ja_JP",
    Region.korea: "ko_KR",
    Platform.korea: "ko_KR",
    Region.latin_america_north: "es_MX",
    Platform.latin_america_north: "es_MX",
    Region.latin_america_south: "es_AR",
    Platform.latin_america_south: "es_AR",
    Region.north_america: "en_US",
    Platform.north_america: "en_US",
    Region.oceania: "en_AU",
    Platform.oceania: "en_AU",
    Region.turkey: "tr_TR",
    Platform.turkey: "tr_TR",
    Region.russia: "ru_RU",
    Platform.russia: "ru_RU",
}


class Key(Enum):
    Q = "Q"
    W = "W"
    E = "E"
    R = "R"


class Resource(Enum):
    mana = "Mana"
    courage = "Courage"
    energy = "Energy"
    fury = "Fury"
    rage = "Rage"
    flow = "Flow"
    ferocity = "Ferocity"
    heat = "Heat"
    shield = "Shield"
    blood_well = "Blood Well"
    crimson_rush = "Crimson Rush"
    none = "None"
    no_cost = "No Cost"


class Side(Enum):
    blue = 100
    red = 200


class GameMode(Enum):
    aram = "ARAM"
    ascension = "ASCENSION"
    classic = "CLASSIC"
    showdown = "FIRSTBLOOD"
    poro_king = "KINGPORO"
    dominion = "ODIN"
    one_for_all = "ONEFORALL"
    tutorial = "TUTORIAL"
    tutorial_1 = "TUTORIAL_MODULE_1"
    tutorial_2 = "TUTORIAL_MODULE_2"
    tutorial_3 = "TUTORIAL_MODULE_3"
    nexus_siege = "SIEGE"
    assassinate = "ASSASSINATE"
    dark_star = "DARKSTAR"
    all_random_summoners_rift = "ARSR"
    urf = "URF"
    doom_bots = "DOOMBOTSTEEMO"
    star_guardian = "STARGUARDIAN"
    project = "PROJECT"
    overcharge = "OVERCHARGE"
    all_random_urf_snow = "SNOWURF"
    practice_tool = "PRACTICETOOL"
    nexus_blitz = "GAMEMODEX"
    odyssey = "ODYSSEY"


class MasteryTree(Enum):
    cunning = "Cunning"
    ferocity = "Ferocity"
    resolve = "Resolve"


class Tier(Enum):
    challenger = "CHALLENGER"
    grandmaster = "GRANDMASTER"
    master = "MASTER"
    diamond = "DIAMOND"
    platinum = "PLATINUM"
    gold = "GOLD"
    silver = "SILVER"
    bronze = "BRONZE"
    iron = "IRON"
    unranked = "UNRANKED"

    def __str__(self):
        return self.name.title()

    @staticmethod
    def _order():
        return {Tier.challenger: 9, Tier.grandmaster: 8, Tier.master: 7,
                Tier.diamond: 6, Tier.platinum: 5, Tier.gold: 4,
                Tier.silver: 3, Tier.bronze: 2, Tier.iron: 1}

    def __lt__(self, other):
        return self._order()[self] < other._order()[other]

    def __gt__(self, other):
        return self._order()[self] > other._order()[other]

    def __le__(self, other):
        return self._order()[self] <= other._order()[other]

    def __ge__(self, other):
        return self._order()[self] >= other._order()[other]


class Division(Enum):
    one = "I"
    two = "II"
    three = "III"
    four = "IV"

    def __str__(self):
        return self.value

    @staticmethod
    def _order():
        return {Division.one: 4, Division.two: 3, Division.three: 2, Division.four: 1}

    def __lt__(self, other):
        return self._order()[self] < other._order()[other]

    def __gt__(self, other):
        return self._order()[self] > other._order()[other]

    def __le__(self, other):
        return self._order()[self] <= other._order()[other]

    def __ge__(self, other):
        return self._order()[self] >= other._order()[other]


class Rank:
    def __init__(self, tier: Tier, division: Division):
        self.tuple = (tier, division)
        self.tier = tier
        self.division = division

    def __str__(self):
        return "<{} {}>".format(self.tuple[0], self.tuple[1])

    def __eq__(self, other):
        return self.tuple == other.tuple

    def __ne__(self, other):
        return self.tuple != other.tuple

    def __lt__(self, other):
        return self.tuple < other.tuple

    def __gt__(self, other):
        return self.tuple > other.tuple

    def __le__(self, other):
        return self.tuple <= other.tuple

    def __ge__(self, other):
        return self.tuple >= other.tuple


class Season(Enum):
    preseason_3 = "PRESEASON3"
    season_3 = "SEASON3"
    preseason_4 = "PRESEASON2014"
    season_4 = "SEASON2014"
    preseason_5 = "PRESEASON2015"
    season_5 = "SEASON2015"
    preseason_6 = "PRESEASON2016"
    season_6 = "SEASON2016"
    preseason_7 = "PRESEASON2017"
    season_7 = "SEASON2017"
    preseason_8 = "PRESEASON2018"
    season_8 = "SEASON2018"
    preseason_9 = "PRESEASON2019"
    season_9 = "SEASON2019"

    @property
    def id(self):
        return SEASON_IDS[self]

    def from_id(id: int):
        return {i: season for season, i in SEASON_IDS.items()}[id]

    def start(self, region: Region) -> arrow.Arrow:
        from .core import Patch
        if Patch._Patch__patches is None:
            Patch.__load__()
        for patch in Patch._Patch__patches[region]:
            if patch.season == self:
                return patch.start

    def end(self, region: Region) -> arrow.Arrow:
        from .core import Patch
        for patch in reversed(Patch._Patch__patches[region]):
            if patch.season == self:
                return patch.end


SEASON_IDS = {
    Season.preseason_3: 0,
    Season.season_3: 1,
    Season.preseason_4: 2,
    Season.season_4: 3,
    Season.preseason_5: 4,
    Season.season_5: 5,
    Season.preseason_6: 6,
    Season.season_6: 7,
    Season.preseason_7: 8,
    Season.season_7: 9,
    Season.preseason_8: 10,
    Season.season_8: 11,
    Season.preseason_9: 12,
    Season.season_9: 13
}


class GameType(Enum):
    custom = "CUSTOM_GAME"
    tutorial = "TUTORIAL_GAME"
    matched = "MATCHED_GAME"


class Lane(Enum):
    top_lane = "TOP_LANE"
    mid_lane = "MID_LANE"
    bot_lane = "BOT_LANE"
    jungle = "JUNGLE"

    def from_match_naming_scheme(string: str):
        return {
            "BOTTOM": Lane.bot_lane,
            "MIDDLE": Lane.mid_lane,
            "MID": Lane.mid_lane,
            "TOP": Lane.top_lane,
            "JUNGLE": Lane.jungle,
            "NONE": None
        }[string]


class Role(Enum):
    duo = "DUO"
    duo_carry = "DUO_CARRY"
    duo_support = "DUO_SUPPORT"
    none = "NONE"
    solo = "SOLO"

    def from_match_naming_scheme(string: str):
        return {
            "DUO": Role.duo,
            "DUO_CARRY": Role.duo_carry,
            "DUO_SUPPORT": Role.duo_support,
            "NONE": Role.none,
            "SOLO": Role.solo
        }[string]


class Position(Enum):
    top = "TOP"
    middle = "MIDDLE"
    jungle = "JUNGLE"
    bottom = "BOTTOM"
    utility = "UTILITY"
    apex = "APEX"
    none = "NONE"

    def from_league_naming_scheme(string: str):
        return {
            "TOP": Position.top,
            "MIDDLE": Position.middle,
            "JUNGLE": Position.jungle,
            "BOTTOM": Position.bottom,
            "UTILITY": Position.support,
            "NONE": Position.none
        }


class SummonersRiftArea(Enum):
    none = "NONE"
    nexus_blue = "NEXUS_BLUE"
    nexus_red = "NEXUS_RED"
    top_lane_blue = "TOP_LANE_BLUE"
    top_lane_purple = "TOP_LANE_PURPLE"
    top_lane_red = "TOP_LANE_RED"
    mid_lane_blue = "MID_LANE_BLUE"
    mid_lane_purple = "MID_LANE_PURPLE"
    mid_lane_red = "MID_LANE_RED"
    bot_lane_blue = "BOT_LANE_BLUE"
    bot_lane_purple = "BOT_LANE_PURPLE"
    bot_lane_red = "BOT_LANE_RED"
    jungle_top_blue = "JUNGLE_TOP_BLUE"
    jungle_top_red = "JUNGLE_TOP_RED"
    jungle_bot_blue = "JUNGLE_BOT_BLUE"
    jungle_bot_red = "JUNGLE_BOT_RED"
    river_top = "RIVER_TOP"
    river_bot = "RIVER_BOT"

    def get_side(self) -> Side:
        if "BLUE" in self.value:
            return Side.blue
        elif "RED" in self.value:
            return Side.red
        else:
            return None

    def get_lane(self) -> Lane:
        if "TOP" in self.value:
            return Lane.top_lane
        elif "MID" in self.value:
            return Lane.mid_lane
        elif "BOT" in self.value:
            return Lane.bot_lane
        elif "JUNGLE" in self.value:
            return Lane.jungle
        else:
            return None

    @staticmethod
    def from_position(position: "Position") -> "SummonersRiftArea":
        from .core.match import Position
        x, y = position.x, position.y

        # Load the map if it isn't already loaded
        try:
            map = SummonersRiftArea.__map
        except AttributeError:
            import os
            from PIL import Image
            script_dir = os.path.dirname(__file__)
            rel_path = './resources/summonersRiftAreas.png'
            map = Image.open(os.path.join(script_dir, rel_path))
            SummonersRiftArea.__map_size = map.size
            map = map.load()
            SummonersRiftArea.__map = map
        image_width, image_height = SummonersRiftArea.__map_size

        min_x = -120
        min_y = -120
        max_x = 14870
        max_y = 14980
        width = max_x - min_x
        height = max_y - min_y
        x = round((x - min_x) / width * (image_width - 1))
        y = round(abs(y - min_y - height) / height * (image_height - 1))
        rgb = map[x, y][0]

        color_mapping = {
            0: SummonersRiftArea.none,
            10: SummonersRiftArea.nexus_blue,
            20: SummonersRiftArea.nexus_red,
            30: SummonersRiftArea.top_lane_blue,
            40: SummonersRiftArea.top_lane_purple,
            50: SummonersRiftArea.top_lane_red,
            60: SummonersRiftArea.mid_lane_blue,
            70: SummonersRiftArea.mid_lane_purple,
            80: SummonersRiftArea.mid_lane_red,
            90: SummonersRiftArea.bot_lane_blue,
            100: SummonersRiftArea.bot_lane_purple,
            110: SummonersRiftArea.bot_lane_red,
            120: SummonersRiftArea.jungle_top_blue,
            130: SummonersRiftArea.jungle_top_red,
            140: SummonersRiftArea.jungle_bot_blue,
            150: SummonersRiftArea.jungle_bot_red,
            160: SummonersRiftArea.river_top,
            170: SummonersRiftArea.river_bot
        }
        return color_mapping.get(rgb, SummonersRiftArea.none)


class Tower(Enum):
    OUTER = "OUTER_TURRET"
    INNER = "INNER_TURRET"
    BASE  = "BASE_TURRET"
    NEXUS = "NEXUS_TURRET"


# References for Queues:
# https://developer.riotgames.com/game-constants.html
# https://discussion.developer.riotgames.com/articles/3482/multiple-queueids-are-being-updated-with-patch-719.html
# https://github.com/stelar7/L4J8/blob/master/src/main/java/no/stelar7/api/l4j8/basic/constants/types/GameQueueType.java
class Queue(Enum):
    custom = "CUSTOM"  # 0
    deprecated_blind_fives = "NORMAL_5x5_BLIND"  # 2
    deprecated_ranked_solo_fives = "CLASSIC"  # 4
    deprecated_ranked_premade_fives = "RANKED_PREMADE_5x5"  # 6
    deprecated_coop_ai_fives = "BOT_5x5"  # 7
    deprecated_blind_threes = "NORMAL_3x3"  # 8
    deprecated_ranked_premade_threes = "RANKED_PREMADE_3x3"  # 9
    deprecated_ranked_flex_threes = "RANKED_FLEX_TT_DEPRECATED"  # 9  # There are two different queue names with ID 9... This one was replaced with queue 470. There is therefore no corresponding queue with ID 9 for this Queue, and instead the Queue with ID 470 will be used when this name is requested, even for very old games. In addition, there are two queues with the name "RANKED_FLEX_TT"; in order to avoid a name conflict, we renamed this one.
    deprecated_draft_fives = "NORMAL_5x5_DRAFT"  # 14
    deprecated_blind_dominion = "ODIN_5x5_BLIND"  # 16
    deprecated_draft_dominion = "ODIN_5x5_DRAFT"  # 17
    deprecated_coop_ai_dominion = "BOT_ODIN_5x5"  # 25
    deprecated_coop_ai_intro_fives = "BOT_5x5_INTRO_DEPRECATED"  # 31  # There are two queues with the name "BOT_5x5_INTRO" so this one has been renamed in order to avoid a conflict.
    deprecated_coop_ai_beginner_fives = "BOT_5x5_BEGINNER_DEPRECATED"  # 32  # There are two queues with the name "BOT_5x5_BEGINNER" so this one has been renamed in order to avoid a conflict.
    deprecated_coop_ai_intermediate_fives = "BOT_5x5_INTERMEDIATE_DEPRECATED"  # 33  # There are two queues with the name "BOT_5x5_INTERMEDIATE" so this one has been renamed in order to avoid a conflict.
    deprecated_ranked_team_threes = "RANKED_TEAM_3x3"  # 41
    deprecated_ranked_team_fives = "RANKED_TEAM_5x5"  # 42
    deprecated_coop_ai_threes = "BOT_TT_3x3"  # 52
    deprecated_team_builder_fives = "GROUP_FINDER_5x5"  # 61
    deprecated_aram = "ARAM_5x5"  # 65
    one_for_all = "ONEFORALL_5x5"  # 70
    showdown_1v1 = "FIRSTBLOOD_1x1"  # 72
    showdown_2v2 = "FIRSTBLOOD_2x2"  # 73
    hexakill_summoners_rift = "SR_6x6"  # 75
    urf = "URF_5x5"  # 76
    mirror_mode_fives = "ONEFORALL_MIRRORMODE_5x5"  # 78
    urf_coop_ai = "BOT_URF_5x5"  # 83
    deprecated_doom_bots_rank_1 = "NIGHTMARE_BOT_5x5_RANK1"  # 91
    deprecated_doom_bots_rank_2 = "NIGHTMARE_BOT_5x5_RANK2"  # 92
    deprecated_doom_bots_rank_5 = "NIGHTMARE_BOT_5x5_RANK5"  # 93
    ascension = "ASCENSION_5x5"  # 96
    hexakill_twisted_treeline = "HEXAKILL"  # 98
    aram_butchers_bridge = "BILGEWATER_ARAM_5x5"  # 100
    deprecated_poro_king = "KING_PORO_5x5"  # 300
    nemesis_draft = "COUNTER_PICK"  # 310
    black_market_brawlers = "BILGEWATER_5x5"  # 313
    deprecated_nexus_siege = "SIEGE"  # 315
    definitely_not_dominion = "DEFINITELY_NOT_DOMINION_5x5"  # 317
    deprecated_all_random_urf = "ARURF_5X5"  # 318
    all_random_summoners_rift = "ARSR_5x5"  # 325
    normal_draft_fives = "TEAM_BUILDER_DRAFT_UNRANKED_5x5"  # 400
    deprecated_ranked_fives = "TEAM_BUILDER_DRAFT_RANKED_5x5"  # 410

    # TODO Evidently we originally had 420 as the commented out queue name below, but it may have changed?
    # TODO But the queue name sent to the Leagues endpoint needs to be RANKED_SOLO_5x5 for ranked solo games.
    ranked_solo_fives = "RANKED_SOLO_5x5"  # 420

    blind_fives = "NORMAL_5V5_BLIND_PICK"  # 430
    ranked_flex_fives = "RANKED_FLEX_SR"  # 440
    aram = "ARAM"  # 450
    blind_threes = "NORMAL_3X3_BLIND_PICK"  # 460
    blood_hunt_assassin = "ASSASSINATE_5x5"  # 600
    dark_star = "DARKSTAR_3x3"  # 610
    ranked_flex_threes = "RANKED_FLEX_TT"  # 470
    clash = "CLASH"  # 700
    coop_ai_intermediate_threes = "BOT_3X3_INTERMEDIATE"  # 800
    coop_ai_intro_threes = "BOT_3X3_INTRO"  # 810
    coop_ai_beginner_threes = "BOT_3X3_BEGINNER"  # 820
    coop_ai_intro_fives = "BOT_5X5_INTRO"  # 830
    coop_ai_beginner_fives = "BOT_5X5_BEGINNER"  # 840
    coop_ai_intermediate_fives = "BOT_5X5_INTERMEDIATE"  # 850
    all_random_urf = "ARURF_5X5"  # 900
    project = "PROJECT"  # 910
    poro_king = "KINGPORO"  # 920
    nexus_siege = "NEXUS_SIEGE"  # 940
    doom_bots_difficult = "NIGHTMARE_BOT_5X5_VOTE"  # 950
    doom_bots = "NIGHTMARE_BOT_5X5"  # 960
    guardian_invasion_normal = "INVASION_NORMAL"  # 980
    guardian_invasion_onslaught = "INVASION_ONSLAUGHT"  # 990
    overcharge = "OVERCHARGE"  # 1000
    all_random_urf_snow = "SNOWURF"  # 1010
    one_for_all_rapid = "ONEFORALL_RAPID_5x5" # 1020
    odyssey_intro = "ODYSSEY_INTRO"  # 1030
    odyssey_cadet = "ODYSSEY_CADET"  # 1040
    odyssey_crewmember = "ODYSSEY_CREWMEMBER"  # 1050
    odyssey_captain = "ODYSSEY_CAPTAIN"  # 1060
    odyssey_onslaught = "ODYSSEY_ONSLAUGHT"  # 1070
    ranked_tft = "RANKED_TFT" # 1100
    normal_tft = "NORMAL_TFT" # 1090
    nexus_blitz = "NEXUS_BLITZ"  # 1200
    tutorial1 = "TUTORIAL_1"  # Summoner's Rift  Tutorial 1
    tutorial2 = "TUTORIAL_2"  # Summoner's Rift  Tutorial 2
    tutorial3 = "TUTORIAL_3"  # Summoner's Rift  Tutorial 3

    def from_id(id: int):
        return {i: season for season, i in QUEUE_IDS.items()}[id]

    @property
    def id(self):
        return QUEUE_IDS[self]


QUEUE_IDS = {
    Queue.custom: 0,  # Custom games
    Queue.deprecated_blind_fives: 2,  # Summoner's Rift    5v5 Blind Pick games    Deprecated in patch 7.19 in favor of queueId 430
    Queue.deprecated_ranked_solo_fives: 4,  # Summoner's Rift    5v5 Ranked Solo games    Deprecated in favor of queueId 420
    Queue.deprecated_ranked_premade_fives: 6,  # Summoner's Rift    5v5 Ranked Premade games    Game mode deprecated
    Queue.deprecated_coop_ai_fives: 7,  # Summoner's Rift    Co-op vs AI games    Deprecated in favor of queueId 32 and 33
    Queue.deprecated_blind_threes: 8,  # Twisted Treeline    3v3 Normal games    Deprecated in patch 7.19 in favor of queueId 460
    Queue.deprecated_ranked_premade_threes: 9,  # Twisted Treeline    3v3 Ranked Flex games    Deprecated in patch 7.19 in favor of queueId 470
    Queue.deprecated_draft_fives: 14,  # Summoner's Rift    5v5 Draft Pick games    Deprecated in favor of queueId 400
    Queue.deprecated_blind_dominion: 16,  # Crystal Scar    5v5 Dominion Blind Pick games    Game mode deprecated
    Queue.deprecated_draft_dominion: 17,  # Crystal Scar    5v5 Dominion Draft Pick games    Game mode deprecated
    Queue.deprecated_coop_ai_dominion: 25,  # Crystal Scar    Dominion Co-op vs AI games    Game mode deprecated
    Queue.deprecated_coop_ai_intro_fives: 31,  # Summoner's Rift    Co-op vs AI Intro Bot games    Deprecated in patch 7.19 in favor of queueId 830
    Queue.deprecated_coop_ai_beginner_fives: 32,  # Summoner's Rift    Co-op vs AI Beginner Bot games    Deprecated in patch 7.19 in favor of queueId 840
    Queue.deprecated_coop_ai_intermediate_fives: 33,  # Summoner's Rift    Co-op vs AI Intermediate Bot games    Deprecated in patch 7.19 in favor of queueId 850
    Queue.deprecated_ranked_team_threes: 41,  # Twisted Treeline    3v3 Ranked Team games    Game mode deprecated
    Queue.deprecated_ranked_team_fives: 42,  # Summoner's Rift    5v5 Ranked Team games    Game mode deprecated
    Queue.deprecated_coop_ai_threes: 52,  # Twisted Treeline    Co-op vs AI games    Deprecated in patch 7.19 in favor of queueId 800
    Queue.deprecated_team_builder_fives: 61,  # Summoner's Rift    5v5 Team Builder games    Game mode deprecated
    Queue.deprecated_aram: 65,  # Howling Abyss    5v5 ARAM games    Deprecated in patch 7.19 in favor of queueId 450
    Queue.one_for_all: 70,  # Summoner's Rift    One for All games
    Queue.showdown_1v1: 72,  # Howling Abyss    1v1 Snowdown Showdown games
    Queue.showdown_2v2: 73,  # Howling Abyss    2v2 Snowdown Showdown games
    Queue.hexakill_summoners_rift: 75,  # Summoner's Rift    6v6 Hexakill games
    Queue.urf: 76,  # Summoner's Rift    Ultra Rapid Fire games
    Queue.mirror_mode_fives: 78,  # Summoner's Rift    Mirrored One for All
    Queue.urf_coop_ai: 83,  # Summoner's Rift    Co-op vs AI Ultra Rapid Fire games
    Queue.deprecated_doom_bots_rank_1: 91,  # Summoner's Rift    Doom Bots Rank 1 games    Deprecated in patch 7.21 in favor of queueId 950
    Queue.deprecated_doom_bots_rank_2: 92,  # Summoner's Rift    Doom Bots Rank 2 games    Deprecated in patch 7.21 in favor of queueId 950
    Queue.deprecated_doom_bots_rank_5: 93,  # Summoner's Rift    Doom Bots Rank 5 games    Deprecated in patch 7.21 in favor of queueId 950
    Queue.ascension: 96,  # Crystal Scar    Ascension games
    Queue.hexakill_twisted_treeline: 98,  # Twisted Treeline    6v6 Hexakill games
    Queue.aram_butchers_bridge: 100,  # Butcher's Bridge    5v5 ARAM games
    Queue.deprecated_poro_king: 300,  # Howling Abyss    King Poro games    Deprecated in patch 7.19 in favor of queueId 920
    Queue.nemesis_draft: 310,  # Summoner's Rift    Nemesis games
    Queue.black_market_brawlers: 313,  # Summoner's Rift    Black Market Brawlers games
    Queue.deprecated_nexus_siege: 315,  # Summoner's Rift    Nexus Siege games    Deprecated in patch 7.19 in favor of queueId 940
    Queue.definitely_not_dominion: 317,  # Crystal Scar    Definitely Not Dominion games
    Queue.deprecated_all_random_urf: 318,  # Summoner's Rift    All Random URF games      Game mode deprecated in patch 8.10 in favor is queueId 900
    Queue.all_random_summoners_rift: 325,  # Summoner's Rift    All Random games
    Queue.normal_draft_fives: 400,  # Summoner's Rift    5v5 Draft Pick games
    Queue.deprecated_ranked_fives: 410,  # Summoner's Rift    5v5 Ranked Dynamic games    Game mode deprecated in patch 6.22
    Queue.ranked_solo_fives: 420,  # Summoner's Rift    5v5 Ranked Solo games
    Queue.blind_fives: 430,  # Summoner's Rift    5v5 Blind Pick games
    Queue.ranked_flex_fives: 440,  # Summoner's Rift    5v5 Ranked Flex games
    Queue.aram: 450,  # Howling Abyss    5v5 ARAM games
    Queue.blind_threes: 460,  # Twisted Treeline    3v3 Blind Pick games
    Queue.ranked_flex_threes: 470,  # Twisted Treeline    3v3 Ranked Flex games
    Queue.blood_hunt_assassin: 600,  # Summoner's Rift    Blood Hunt Assassin games
    Queue.dark_star: 610,  # Cosmic Ruins    Dark Star games
    Queue.clash: 700,  # Summoner's Rift    Clash games
    Queue.coop_ai_intermediate_threes: 800,  # Twisted Treeline    Co-op vs. AI Intermediate Bot games
    Queue.coop_ai_intro_threes: 810,  # Twisted Treeline    Co-op vs. AI Intro Bot games
    Queue.coop_ai_beginner_threes: 820,  # Twisted Treeline    Co-op vs. AI Beginner Bot games
    Queue.coop_ai_intro_fives: 830,  # Summoner's Rift    Co-op vs. AI Intro Bot games
    Queue.coop_ai_beginner_fives: 840,  # Summoner's Rift    Co-op vs. AI Beginner Bot games
    Queue.coop_ai_intermediate_fives: 850,  # Summoner's Rift    Co-op vs. AI Intermediate Bot games
    Queue.all_random_urf: 900,  # Summoner's Rift    All Random URF games
    Queue.project: 910,
    Queue.poro_king: 920,  # Howling Abyss    Legend of the Poro King
    Queue.nexus_siege: 940,  # Summoner's Rift    Nexus Siege games
    Queue.doom_bots_difficult: 950,  # Summoner's Rift    Doom Bots games /w difficulty voting
    Queue.doom_bots: 960,  # Summoner's Rift    Doom Bots games
    Queue.guardian_invasion_normal: 980,  # Valoran City Park    Star Guardian Invasion: Normal games
    Queue.guardian_invasion_onslaught: 990,  # Valoran City Park    Star Guardian Invasion: Onslaught games
    Queue.overcharge: 1000,  # Overcharge, PROJECT: Hunters games
    Queue.all_random_urf_snow: 1010,  # Summoner's Rift, Snow ARURF games
    Queue.one_for_all_rapid: 1020, # Summoner's Rift  One for All games (increased gold and exp gain)
    Queue.odyssey_intro: 1030,  # Odyssey: Extraction
    Queue.odyssey_cadet: 1040,  # Odyssey: Extraction
    Queue.odyssey_crewmember: 1050,  # Odyssey: Extraction
    Queue.odyssey_captain: 1060,  # Odyssey: Extraction
    Queue.odyssey_onslaught: 1070,  # Odyssey: Extraction
    Queue.ranked_tft: 1100, #  Convergence, Ranked Teamfight Tactics games
    Queue.normal_tft: 1090, #  Convergence, Normal Teamfight Tactics games
    Queue.nexus_blitz: 1200,  # Nexus Blitz map    Nexus Blitz
    Queue.tutorial1: 2000,  # Summoner's Rift  Tutorial 1
    Queue.tutorial2: 2010,  # Summoner's Rift  Tutorial 2
    Queue.tutorial3: 2020,  # Summoner's Rift  Tutorial 3
}

RANKED_QUEUES = {
    Queue.deprecated_ranked_solo_fives,  # Summoner's Rift    5v5 Ranked Solo games    Deprecated in favor of queueId 420
    Queue.deprecated_ranked_premade_fives,  # Summoner's Rift    5v5 Ranked Premade games    Game mode deprecated
    Queue.deprecated_ranked_premade_threes,  # Twisted Treeline    3v3 Ranked Flex games    Deprecated in patch 7.19 in favor of queueId 470
    Queue.deprecated_ranked_team_threes,  # Twisted Treeline    3v3 Ranked Team games    Game mode deprecated
    Queue.deprecated_ranked_team_fives,  # Summoner's Rift    5v5 Ranked Team games    Game mode deprecated
    Queue.deprecated_ranked_fives,  # Summoner's Rift    5v5 Ranked Dynamic games    Game mode deprecated in patch 6.22
    Queue.ranked_solo_fives,  # Summoner's Rift    5v5 Ranked Solo games
    Queue.ranked_flex_fives,  # Summoner's Rift    5v5 Ranked Flex games
    Queue.ranked_flex_threes,  # Twisted Treeline    3v3 Ranked Flex games
    Queue.ranked_tft, # Convergence  Ranked Teamfight Tactics games
}



from typing import Type, Mapping, Any, Iterable, MutableMapping, TypeVar, Tuple, Callable, Generator
import datetime
import copy

from datapipelines import DataSource, DataSink, PipelineContext, Query, validate_query, NotFoundError
from backends.cache_backend import DjangoCacheBackend

from ..data import Platform, Region, Queue
from ..dto.common import DtoObject
from ..dto.champion import ChampionRotationDto
from ..dto.championmastery import ChampionMasteryDto, ChampionMasteryListDto
from ..dto.league import MasterLeagueListDto, ChallengerLeagueListDto, GrandmasterLeagueListDto
from ..dto.staticdata import ChampionDto, ChampionListDto, RuneDto, RuneListDto, ItemDto, ItemListDto, SummonerSpellDto, SummonerSpellListDto, MapDto, MapListDto, RealmDto, ProfileIconDataDto, ProfileIconDetailsDto, LanguagesDto, LanguageStringsDto, VersionListDto
from ..dto.match import MatchDto, TimelineDto, MatchListDto
from ..dto.summoner import SummonerDto
from ..dto.status import ShardStatusDto
from ..dto.spectator import CurrentGameInfoDto, FeaturedGamesDto
from ..dto.patch import PatchListDto
from ..datastores.uniquekeys import convert_region_to_platform
from ..datastores.riotapi.common import _get_default_locale, _get_latest_version
from logging import getLogger

LOGGER = getLogger(__name__)

T = TypeVar("T")

default_expirations = {
    RealmDto: datetime.timedelta(hours=6),
    VersionListDto: datetime.timedelta(hours=6),
    ChampionDto: datetime.timedelta(days=1),
    ChampionListDto: datetime.timedelta(days=1),
    RuneDto: datetime.timedelta(days=1),
    RuneListDto: datetime.timedelta(days=1),
    ItemDto: datetime.timedelta(days=1),
    ItemListDto: datetime.timedelta(days=1),
    SummonerSpellDto: datetime.timedelta(days=1),
    SummonerSpellListDto: datetime.timedelta(days=1),
    MapDto: datetime.timedelta(days=7),
    MapListDto: datetime.timedelta(days=7),
    ProfileIconDetailsDto: datetime.timedelta(days=1),
    ProfileIconDataDto: datetime.timedelta(days=1),
    LanguagesDto: datetime.timedelta(days=7),
    LanguageStringsDto: datetime.timedelta(days=7),
    ChampionRotationDto: datetime.timedelta(hours=6),
    ChampionMasteryDto: datetime.timedelta(hours=6),
    ChampionMasteryListDto: datetime.timedelta(hours=6),
    ChallengerLeagueListDto: datetime.timedelta(hours=6),
    GrandmasterLeagueListDto: datetime.timedelta(hours=6),
    MasterLeagueListDto: datetime.timedelta(hours=6),
    MatchDto: datetime.timedelta(days=7),
    TimelineDto: datetime.timedelta(days=7),
    SummonerDto: datetime.timedelta(minutes=15),
    ShardStatusDto: datetime.timedelta(minutes=15),
    CurrentGameInfoDto: datetime.timedelta(minutes=15),
    FeaturedGamesDto: datetime.timedelta(minutes=15),
    PatchListDto: datetime.timedelta(days=1)
}


class DjangoCache(DataSource, DataSink):
    def __init__(self, expirations: Mapping[type, float] = None, alias: str = None, logs_enabled: bool = True) -> None:
        self._alias = alias
        self._cache = DjangoCacheBackend(alias, logs_enabled)
        self._expirations = dict(expirations) if expirations is not None else default_expirations
        #Not handling cache name check as Django will do it.
        for key, value in list(self._expirations.items()):
            if isinstance(key, str):
                if not key.endswith("Dto"):
                    raise NotImplementedError("Django's cache framework cannot cache objects of type 'cassiopeia.core'")
                new_key = globals()[key]
                self._expirations[new_key] = self._expirations.pop(key)
                key = new_key
            if value != -1 and isinstance(value, datetime.timedelta):
                self._expirations[key] = value.total_seconds()
            elif value == -1:
                LOGGER.warn(f"[Traceback: django_cassiopeia > {self._alias}] WARNING: {key.__name__} has set with 'forever' caching, and a cache is meant to expire at some point.")

    @DataSource.dispatch
    def get(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: Mapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    @DataSink.dispatch
    def put(self, type: Type[T], item: T, context: PipelineContext = None) -> None:
        pass

    @DataSink.dispatch
    def put_many(self, type: Type[T], items: Iterable[T], context: PipelineContext = None) -> None:
        pass

    def _get(self, key: str) -> T:
        try:
            return self._cache.get(key)
        except KeyError:
            raise NotFoundError

    def _put(self, key: str, item: DtoObject) -> None:
        expire_seconds = self._expirations.get(item.__class__, default_expirations[item.__class__])
        if expire_seconds != 0:
            self._cache.put(key, item, expire_seconds)

    def _put_many(self, pairs: Mapping[str, DtoObject], clsname: DtoObject) -> None:
        expire_seconds = self._expirations.get(clsname, default_expirations[clsname])
        if expire_seconds != 0:
            self._cache.put_many(pairs, expire_seconds)

    def clear(self, type: Type[T] = None):
        self._cache.clear(type.__name__)

    # This does nothing, Django is handling expiration.
    def expire(self, type: Type[T] = None):
        self._cache.expire(type)

    ######################
    # Champion Masteries #
    ######################

    _validate_get_champion_mastery_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(str).also. \
        has("champion.id").as_(int)

    @get.register(ChampionMasteryDto)
    @validate_query(_validate_get_champion_mastery_query, convert_region_to_platform)
    def get_champion_mastery(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionMasteryDto:
        champions_query = copy.deepcopy(query)
        champions_query.pop("champion.id")
        try:
            champions = self.get_champion_mastery_list(query=champions_query, context=context)
        except NotFoundError:
            raise NotFoundError

        def find_matching_attribute(list_of_dtos, attrname, attrvalue):
            for dto in list_of_dtos:
                if dto.get(attrname, None) == attrvalue:
                    return dto

        champion = find_matching_attribute(champions["masteries"], "championId", query["champion.id"])
        if champion is None:
            raise NotFoundError
        return ChampionMasteryDto(champion)

    _validate_get_champion_mastery_list_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(str)

    @get.register(ChampionMasteryListDto)
    @validate_query(_validate_get_champion_mastery_list_query, convert_region_to_platform)
    def get_champion_mastery_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionMasteryListDto:
        self._validate_get_champion_mastery_list_query(query, context)
        platform = query["platform"].value
        summoner_id = query["summoner.id"]
        key = "{clsname}.{platform}.{summoner_id}".format(clsname=ChampionMasteryListDto.__name__,
                                                           platform=platform,
                                                           summoner_id=summoner_id)
        return ChampionMasteryListDto(self._get(key))

    @put.register(ChampionMasteryListDto)
    def put_champion_mastery_list(self, item: ChampionMasteryListDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        summoner_id = item["summonerId"]
        key = "{clsname}.{platform}.{summoner_id}".format(clsname=ChampionMasteryListDto.__name__,
                                                           platform=platform,
                                                           summoner_id=summoner_id)
        self._put(key, item)

    #############
    # Champions #
    #############

    _validate_get_champion_status_list_query = Query. \
        has("platform").as_(Platform)

    @get.register(ChampionRotationDto)
    @validate_query(_validate_get_champion_status_list_query, convert_region_to_platform)
    def get_champion_status_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionRotationDto:
        platform = query["platform"].value
        key = "{clsname}.{platform}".format(clsname="ChampionRotationDto",
                                                           platform=platform)
        return ChampionRotationDto(self._get(key))

    @put.register(ChampionRotationDto)
    def put_champion_status_list(self, item: ChampionRotationDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}".format(clsname="ChampionRotationDto",
                                                           platform=platform)
        self._put(key, item)

    ###########
    # Leagues #
    ###########

    # Challenger

    _validate_get_challenger_league_query = Query. \
        has("platform").as_(Platform).also. \
        has("queue").as_(Queue)

    @get.register(ChallengerLeagueListDto)
    @validate_query(_validate_get_challenger_league_query, convert_region_to_platform)
    def get_challenger_league(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChallengerLeagueListDto:
        key = "{clsname}.{platform}.{queue}".format(clsname=ChallengerLeagueListDto.__name__,
                                                    platform=query["platform"].value,
                                                    queue=query["queue"].value)
        return ChallengerLeagueListDto(self._get(key))

    @put.register(ChallengerLeagueListDto)
    def put_challenger_league(self, item: ChallengerLeagueListDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}.{queue}".format(clsname=ChallengerLeagueListDto.__name__,
                                                    platform=platform,
                                                    queue=item["queue"])
        self._put(key, item)

    # Grandmaster

    _validate_get_grandmaster_league_query = Query. \
        has("platform").as_(Platform).also. \
        has("queue").as_(Queue)

    @get.register(GrandmasterLeagueListDto)
    @validate_query(_validate_get_grandmaster_league_query, convert_region_to_platform)
    def get_grandmaster_league(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> GrandmasterLeagueListDto:
        key = "{clsname}.{platform}.{queue}".format(clsname=GrandmasterLeagueListDto.__name__,
                                                    platform=query["platform"].value,
                                                    queue=query["queue"].value)
        return GrandmasterLeagueListDto(self._get(key))

    @put.register(GrandmasterLeagueListDto)
    def put_grandmaster_league(self, item: GrandmasterLeagueListDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}.{queue}".format(clsname=GrandmasterLeagueListDto.__name__,
                                                    platform=platform,
                                                    queue=item["queue"])
        self._put(key, item)

    # Master

    _validate_get_master_league_query = Query. \
        has("platform").as_(Platform).also. \
        has("queue").as_(Queue)

    @get.register(MasterLeagueListDto)
    @validate_query(_validate_get_master_league_query, convert_region_to_platform)
    def get_master_league(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MasterLeagueListDto:
        key = "{clsname}.{platform}.{queue}".format(clsname=MasterLeagueListDto.__name__,
                                                    platform=query["platform"].value,
                                                    queue=query["queue"].value)
        return MasterLeagueListDto(self._get(key))

    @put.register(MasterLeagueListDto)
    def put_master_league(self, item: MasterLeagueListDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}.{queue}".format(clsname=MasterLeagueListDto.__name__,
                                                    platform=platform,
                                                    queue=item["queue"])
        self._put(key, item)

    #########
    # Match #
    #########

    # Match

    _validate_get_match_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform)

    @get.register(MatchDto)
    @validate_query(_validate_get_match_query, convert_region_to_platform)
    def get_match(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MatchDto:
        key = "{clsname}.{platform}.{id}".format(clsname=MatchDto.__name__,
                                                 platform=query["platform"].value,
                                                 id=query["id"])
        return MatchDto(self._get(key))

    @put.register(MatchDto)
    def put_match(self, item: MatchDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}.{id}".format(clsname=MatchDto.__name__,
                                                 platform=platform,
                                                 id=item["gameId"])
        self._put(key, item)

    # Match list

    # This is cool and useful functionality, but it really only works if we can pull the entire match
    # history in one go. For now, we just won't save the match history to disk at all.
    """
    _validate_get_match_list_query = Query. \
        has("account.id").as_(int).also. \
        has("platform").as_(Platform).also. \
        can_have("beginTime").as_(int).also. \
        can_have("endTime").as_(int).also. \
        can_have("beginIndex").as_(int).also. \
        can_have("endIndex").as_(int).also. \
        can_have("seasons").as_(Iterable).also. \
        can_have("champion.ids").as_(Iterable).also. \
        can_have("queues").as_(Iterable).also. \
        can_have("forceRefresh").with_default(False)
    @get.register(MatchListDto)
    def get_match_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MatchListDto:
        MatchDiskService._validate_get_match_list_query(query, context)
        platform = query["platform"].value
        queues = "|".join(sorted({queue.value for queue in query.get("queues", {})}))
        seasons = "|".join(sorted({season.value for season in query.get("seasons", {})}))
        champions = "|".join(sorted({champion.value for champion in query.get("champion.ids", {})}))
        key = "{clsname}.{platform}.{account_id}.{queues}.{seasons}.{champions}".format(clsname=MatchListDto.__name__,
                                                                                        platform=platform,
                                                                                        account_id=query["account.id"],
                                                                                        queues=queues,
                                                                                        seasons=seasons,
                                                                                        champions=champions)
        data = self._get(key)
        # The above line will throw a NotFoundError if the matchlist for this summoner doesn't exist at all.
        # However, if it does exist, let's try to be smart about pulling the remaining data.
        most_recent = data["matches"][0]["timestamp"]
        # Choose 30 minutes to refresh match history
        refresh_from_expiration = datetime.datetime.fromtimestamp(most_recent/1000) < datetime.datetime.now() - datetime.timedelta(minutes=30)
        if query["forceRefresh"] or refresh_from_expiration:
            new_query = copy.deepcopy(query)
            new_query.pop("beginTime", None)
            new_query.pop("endTime", None)
            new_query.pop("beginIndex", None)
            new_query.pop("endIndex", None)
            new_query["beginTime"] = most_recent + 1  # Add 1 ms so we don't get the last game we have.
            new_data = context[context.Keys.PIPELINE].get(MatchListDto, query=new_query, context=context)
            data["matches"].extend(new_data["matches"])
        return MatchListDto(self._get(key))
    @put.register(MatchListDto)
    def put_match_list(self, item: MatchListDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        queues = "|".join(sorted({queue.value for queue in item["queue"]}))
        seasons = "|".join(sorted({season.value for season in item["season"]}))
        champions = "|".join(sorted({champion.value for champion in item["champion"]}))
        key = "{clsname}.{platform}.{account_id}.{queues}.{seasons}.{champions}".format(clsname=MatchListDto.__name__,
                                                                                        platform=platform,
                                                                                        account_id=item["accountId"],
                                                                                        queues=queues,
                                                                                        seasons=seasons,
                                                                                        champions=champions)
        self._put(key, item)
    """

    # Timeline

    _validate_get_timeline_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform)

    @get.register(TimelineDto)
    @validate_query(_validate_get_timeline_query, convert_region_to_platform)
    def get_timeline(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> TimelineDto:
        key = "{clsname}.{platform}.{id}".format(clsname=TimelineDto.__name__,
                                                 platform=query["platform"].value,
                                                 id=query["id"])
        return TimelineDto(self._get(key))

    @put.register(TimelineDto)
    def put_timeline(self, item: TimelineDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}.{id}".format(clsname=TimelineDto.__name__,
                                                 platform=platform,
                                                 id=item["matchId"])
        self._put(key, item)

    #########
    # Patch #
    #########

    @get.register(PatchListDto)
    def get_patches(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> PatchListDto:
        key = "{clsname}".format(clsname=PatchListDto.__name__)
        return PatchListDto(self._get(key))

    @put.register(PatchListDto)
    def put_patches(self, item: PatchListDto, context: PipelineContext = None) -> None:
        key = "{clsname}".format(clsname=PatchListDto.__name__)
        self._put(key, item)

    ##################
    # Featured Games #
    ##################

    _validate_get_featured_games_query = Query. \
        has("platform").as_(Platform)

    @get.register(FeaturedGamesDto)
    @validate_query(_validate_get_featured_games_query, convert_region_to_platform)
    def get_featured_games(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> FeaturedGamesDto:
        key = "{clsname}.{platform}".format(clsname=FeaturedGamesDto.__name__, platform=query["platform"].value)
        return FeaturedGamesDto(self._get(key))

    @put.register(FeaturedGamesDto)
    def put_featured_games(self, item: FeaturedGamesDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}".format(clsname=FeaturedGamesDto.__name__, platform=platform)
        self._put(key, item)

    ################
    # Current Game #
    ################

    _validate_get_current_game_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(str)

    @get.register(CurrentGameInfoDto)
    @validate_query(_validate_get_current_game_query, convert_region_to_platform)
    def get_current_game(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> CurrentGameInfoDto:
        key = "{clsname}.{platform}.{id}".format(clsname=CurrentGameInfoDto.__name__,
                                                 platform=query["platform"].value,
                                                 id=query["summoner.id"])
        return CurrentGameInfoDto(self._get(key))

    @put.register(CurrentGameInfoDto)
    def put_current_game(self, item: CurrentGameInfoDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}.{id}".format(clsname=CurrentGameInfoDto.__name__,
                                                 platform=platform,
                                                 id=item["summonerId"])
        self._put(key, item)

    ############
    # Versions #
    ############

    _validate_get_versions_query = Query. \
        has("platform").as_(Platform)

    @get.register(VersionListDto)
    @validate_query(_validate_get_versions_query, convert_region_to_platform)
    def get_versions(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> VersionListDto:
        key = "{clsname}.{platform}".format(clsname=VersionListDto.__name__, platform=query["platform"].value)
        return VersionListDto(self._get(key))

    @put.register(VersionListDto)
    def put_versions(self, item: VersionListDto, context: PipelineContext = None) -> None:
        key = "{clsname}.{platform}".format(clsname=VersionListDto.__name__, platform=Region(item["region"]).platform.value)
        self._put(key, item)

    ##########
    # Realms #
    ##########

    _validate_get_realms_query = Query. \
        has("platform").as_(Platform)

    @get.register(RealmDto)
    @validate_query(_validate_get_realms_query, convert_region_to_platform)
    def get_realms(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> RealmDto:
        key = "{clsname}.{platform}".format(clsname=RealmDto.__name__, platform=query["platform"].value)
        return RealmDto(self._get(key))

    @put.register(RealmDto)
    def put_realms(self, item: RealmDto, context: PipelineContext = None) -> None:
        key = "{clsname}.{platform}".format(clsname=RealmDto.__name__, platform=Region(item["region"]).platform.value)
        self._put(key, item)

    #############
    # Champions #
    #############

    _validate_get_champion_query = Query. \
        has("id").as_(int).or_("name").as_(str).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(ChampionDto)
    @validate_query(_validate_get_champion_query, convert_region_to_platform)
    def get_champion(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionDto:
        champions_query = copy.deepcopy(query)
        if "id" in champions_query:
            champions_query.pop("id")
        if "name" in champions_query:
            champions_query.pop("name")
        champions = context[context.Keys.PIPELINE].get(ChampionListDto, query=champions_query)

        def find_matching_attribute(list_of_dtos, attrname, attrvalue):
            for dto in list_of_dtos:
                if dto.get(attrname, None) == attrvalue:
                    return dto

        if "id" in query:
            champion = find_matching_attribute(champions["data"].values(), "id", query["id"])
        elif "name" in query:
            champion = find_matching_attribute(champions["data"].values(), "name", query["name"])
        else:
            raise ValueError("Impossible!")
        if champion is None:
            raise NotFoundError
        champion["region"] = query["platform"].region.value
        champion["version"] = query["version"]
        champion["locale"] = query["locale"]
        champion["includedData"] = query["includedData"]
        return ChampionDto(champion)

    _validate_get_champion_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(ChampionListDto)
    @validate_query(_validate_get_champion_list_query, convert_region_to_platform)
    def get_champion_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionListDto:
        platform = query["platform"].value
        version = query["version"]
        locale = query["locale"]
        included_data = "|".join(sorted(query["includedData"]))
        key = "{clsname}.{platform}.{version}.{locale}.{included_data}".format(clsname=ChampionListDto.__name__,
                                                                                            platform=platform,
                                                                                            version=version,
                                                                                            locale=locale,
                                                                                            included_data=included_data)
        data = self._get(key)
        data["data"] = {key: ChampionDto(champion) for key, champion in data["data"].items()}
        return ChampionListDto(data)

    @put.register(ChampionListDto)
    def put_champion_list(self, item: ChampionListDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        included_data = "|".join(sorted(item["includedData"]))
        key = "{clsname}.{platform}.{version}.{locale}.{included_data}".format(clsname=ChampionListDto.__name__,
                                                                                            platform=platform,
                                                                                            version=item["version"],
                                                                                            locale=item["locale"],
                                                                                            included_data=included_data)
        self._put(key, item)

    #########
    # Items #
    #########

    _validate_get_item_query = Query. \
        has("id").as_(int).or_("name").as_(str).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(ItemDto)
    @validate_query(_validate_get_item_query, convert_region_to_platform)
    def get_item(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ItemDto:
        items_query = copy.deepcopy(query)
        if "id" in items_query:
            items_query.pop("id")
        if "name" in items_query:
            items_query.pop("name")
        items = context[context.Keys.PIPELINE].get(ItemListDto, query=items_query)

        def find_matching_attribute(list_of_dtos, attrname, attrvalue):
            for dto in list_of_dtos:
                if dto.get(attrname, None) == attrvalue:
                    return dto

        if "id" in query:
            item = find_matching_attribute(items["data"].values(), "id", query["id"])
        elif "name" in query:
            item = find_matching_attribute(items["data"].values(), "name", query["name"])
        else:
            raise ValueError("Impossible!")
        if item is None:
            raise NotFoundError
        item["region"] = query["platform"].region.value
        item["version"] = query["version"]
        item["locale"] = query["locale"]
        item["includedData"] = query["includedData"]
        return ItemDto(item)

    _validate_get_item_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(ItemListDto)
    @validate_query(_validate_get_item_list_query, convert_region_to_platform)
    def get_item_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ItemListDto:
        platform = query["platform"].value
        version = query["version"]
        locale = query["locale"]
        included_data = "|".join(sorted(query["includedData"]))
        key = "{clsname}.{platform}.{version}.{locale}.{included_data}".format(clsname=ItemListDto.__name__,
                                                                               platform=platform,
                                                                               version=version,
                                                                               locale=locale,
                                                                               included_data=included_data)
        data = self._get(key)
        for key, item in data["data"].items():
            item = ItemDto(item)
            data["data"][key] = item
        return ItemListDto(data)

    @put.register(ItemListDto)
    def put_item_list(self, item: ItemListDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        included_data = "|".join(sorted(item["includedData"]))
        key = "{clsname}.{platform}.{version}.{locale}.{included_data}".format(clsname=ItemListDto.__name__,
                                                                               platform=platform,
                                                                               version=item["version"],
                                                                               locale=item["locale"],
                                                                               included_data=included_data)
        self._put(key, item)

    ##################
    # SummonerSpells #
    ##################

    _validate_get_summoner_spell_query = Query. \
        has("id").as_(int).or_("name").as_(str).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(SummonerSpellDto)
    @validate_query(_validate_get_summoner_spell_query, convert_region_to_platform)
    def get_summoner_spell(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> SummonerSpellDto:
        summoner_spells_query = copy.deepcopy(query)
        if "id" in summoner_spells_query:
            summoner_spells_query.pop("id")
        if "name" in summoner_spells_query:
            summoner_spells_query.pop("name")
        summoner_spells = context[context.Keys.PIPELINE].get(SummonerSpellListDto, query=summoner_spells_query)

        def find_matching_attribute(list_of_dtos, attrname, attrvalue):
            for dto in list_of_dtos:
                if dto.get(attrname, None) == attrvalue:
                    return dto

        if "id" in query:
            summoner_spell = find_matching_attribute(summoner_spells["data"].values(), "id", query["id"])
        elif "name" in query:
            summoner_spell = find_matching_attribute(summoner_spells["data"].values(), "name", query["name"])
        else:
            raise ValueError("Impossible!")
        if summoner_spell is None:
            raise NotFoundError
        summoner_spell["region"] = query["platform"].region.value
        summoner_spell["version"] = query["version"]
        summoner_spell["locale"] = query["locale"]
        summoner_spell["includedData"] = query["includedData"]
        return SummonerSpellDto(summoner_spell)

    _validate_get_summoner_spell_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(SummonerSpellListDto)
    @validate_query(_validate_get_summoner_spell_list_query, convert_region_to_platform)
    def get_summoner_spell_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> SummonerSpellListDto:
        platform = query["platform"].value
        version = query["version"]
        locale = query["locale"]
        included_data = "|".join(sorted(query["includedData"]))
        key = "{clsname}.{platform}.{version}.{locale}.{included_data}".format(clsname=SummonerSpellListDto.__name__,
                                                                               platform=platform,
                                                                               version=version,
                                                                               locale=locale,
                                                                               included_data=included_data)
        return SummonerSpellListDto(self._get(key))

    @put.register(SummonerSpellListDto)
    def put_summoner_spell_list(self, item: SummonerSpellListDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        included_data = "|".join(sorted(item["includedData"]))
        key = "{clsname}.{platform}.{version}.{locale}.{included_data}".format(clsname=SummonerSpellListDto.__name__,
                                                                               platform=platform,
                                                                               version=item["version"],
                                                                               locale=item["locale"],
                                                                               included_data=included_data)
        self._put(key, item)

    ########
    # Maps #
    ########

    _validate_get_map_query = Query. \
        has("id").as_(int).or_("name").as_(str).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    @get.register(MapDto)
    @validate_query(_validate_get_map_query, convert_region_to_platform)
    def get_map(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MapDto:
        maps_query = copy.deepcopy(query)
        if "id" in maps_query:
            maps_query.pop("id")
        if "name" in maps_query:
            maps_query.pop("name")
        maps = context[context.Keys.PIPELINE].get(MapListDto, query=maps_query)

        def find_matching_attribute(list_of_dtos, attrname, attrvalue):
            for dto in list_of_dtos:
                if dto.get(attrname, None) == attrvalue:
                    return dto

        if "id" in query:
            map = find_matching_attribute(maps["data"].values(), "mapId", str(query["id"]))
        elif "name" in query:
            map = find_matching_attribute(maps["data"].values(), "mapName", query["name"])
        else:
            raise ValueError("Impossible!")
        if map is None:
            raise NotFoundError
        map["region"] = query["platform"].region.value
        map["version"] = query["version"]
        map["locale"] = query["locale"]
        return MapDto(map)

    _validate_get_map_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    @get.register(MapListDto)
    @validate_query(_validate_get_map_list_query, convert_region_to_platform)
    def get_map_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MapListDto:
        platform = query["platform"].value
        version = query["version"]
        locale = query["locale"]
        key = "{clsname}.{platform}.{version}.{locale}".format(clsname=MapListDto.__name__,
                                                               platform=platform,
                                                               version=version,
                                                               locale=locale)
        return MapListDto(self._get(key))

    @put.register(MapListDto)
    def put_map_list(self, item: MapListDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}.{version}.{locale}".format(clsname=MapListDto.__name__,
                                                               platform=platform,
                                                               version=item["version"],
                                                               locale=item["locale"])
        self._put(key, item)

    #################
    # Profile Icons #
    #################

    _validate_get_profile_icons_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    @get.register(ProfileIconDataDto)
    @validate_query(_validate_get_profile_icons_query, convert_region_to_platform)
    def get_profile_icons(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ProfileIconDataDto:
        platform = query["platform"].value
        version = query["version"]
        locale = query["locale"]
        key = "{clsname}.{platform}.{version}.{locale}".format(clsname=ProfileIconDataDto.__name__,
                                                               platform=platform,
                                                               version=version,
                                                               locale=locale)
        return ProfileIconDataDto(self._get(key))

    @put.register(ProfileIconDataDto)
    def put_profile_icons(self, item: ProfileIconDataDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}.{version}.{locale}".format(clsname=ProfileIconDataDto.__name__,
                                                               platform=platform,
                                                               version=item["version"],
                                                               locale=item["locale"])
        self._put(key, item)

    ############
    # Language #
    ############

    _validate_get_languages_query = Query. \
        has("platform").as_(Platform)

    @get.register(LanguagesDto)
    @validate_query(_validate_get_languages_query, convert_region_to_platform)
    def get_language(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LanguagesDto:
        platform = query["platform"].value
        key = "{clsname}.{platform}".format(clsname=LanguagesDto.__name__, platform=platform)
        return LanguagesDto(self._get(key))

    _validate_get_many_languages_query = Query. \
        has("platforms").as_(Iterable)

    @put.register(LanguagesDto)
    def put_language(self, item: LanguagesDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}".format(clsname=LanguagesDto.__name__, platform=platform)
        self._put(key, item)

    ####################
    # Language Strings #
    ####################

    _validate_get_language_strings_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    @get.register(LanguageStringsDto)
    @validate_query(_validate_get_language_strings_query, convert_region_to_platform)
    def get_language_strings(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LanguageStringsDto:
        platform = query["platform"].value
        version = query["version"]
        locale = query["locale"]
        key = "{clsname}.{platform}.{version}.{locale}".format(clsname=LanguageStringsDto.__name__,
                                                               platform=platform,
                                                               version=version,
                                                               locale=locale)
        return LanguageStringsDto(self._get(key))

    _validate_get_many_language_strings_query = Query. \
        has("platforms").as_(Iterable).also. \
        can_have("version").as_(str).also. \
        can_have("locale").as_(str)

    @put.register(LanguageStringsDto)
    def put_language_strings(self, item: LanguageStringsDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}.{version}.{locale}".format(clsname=LanguageStringsDto.__name__,
                                                               platform=platform,
                                                               version=item["version"],
                                                               locale=item["locale"])
        self._put(key, item)

    #########
    # Runes #
    #########

    _validate_get_rune_query = Query. \
        has("id").as_(int).or_("name").as_(str).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    @get.register(RuneDto)
    @validate_query(_validate_get_rune_query, convert_region_to_platform)
    def get_rune(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> RuneDto:
        runes_query = copy.deepcopy(query)
        if "id" in runes_query:
            runes_query.pop("id")
        if "name" in runes_query:
            runes_query.pop("name")
        runes = context[context.Keys.PIPELINE].get(RuneListDto, query=runes_query)

        def find_matching_attribute(list_of_dtos, attrname, attrvalue):
            for dto in list_of_dtos:
                if dto.get(attrname, None) == attrvalue:
                    return dto

        if "id" in query:
            rune = find_matching_attribute(runes["data"], "runeId", str(query["id"]))
        elif "name" in query:
            rune = find_matching_attribute(runes["data"], "runeName", query["name"])
        else:
            raise ValueError("Impossible!")
        if rune is None:
            raise NotFoundError
        rune["region"] = query["platform"].region.value
        rune["version"] = query["version"]
        rune["locale"] = query["locale"]
        return RuneDto(rune)

    _validate_get_rune_list_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    @get.register(RuneListDto)
    @validate_query(_validate_get_rune_list_query, convert_region_to_platform)
    def get_rune_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> RuneListDto:
        platform = query["platform"].value
        version = query["version"]
        locale = query["locale"]
        key = "{clsname}.{platform}.{version}.{locale}".format(clsname=RuneListDto.__name__,
                                                               platform=platform,
                                                               version=version,
                                                               locale=locale)
        return RuneListDto(self._get(key))

    @put.register(RuneListDto)
    def put_rune_list(self, item: RuneListDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}.{version}.{locale}".format(clsname=RuneListDto.__name__,
                                                               platform=platform,
                                                               version=item["version"],
                                                               locale=item["locale"])
        self._put(key, item)

    ##########
    # Status #
    ##########

    _validate_get_status_query = Query. \
        has("platform").as_(Platform)

    @get.register(ShardStatusDto)
    @validate_query(_validate_get_status_query, convert_region_to_platform)
    def get_status(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ShardStatusDto:
        key = "{clsname}.{platform}".format(clsname=ShardStatusDto.__name__, platform=query["platform"].value)
        return ShardStatusDto(self._get(key))

    @put.register(ShardStatusDto)
    def put_status(self, item: ShardStatusDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        key = "{clsname}.{platform}".format(clsname=ShardStatusDto.__name__, platform=platform)
        self._put(key, item)

    ############
    # Summoner #
    ############

    _validate_get_summoner_query = Query. \
        has("id").as_(str). \
        or_("accountId").as_(str). \
        or_("puuid").as_(str). \
        or_("name").as_(str).also. \
        has("platform").as_(Platform)

    @get.register(SummonerDto)
    @validate_query(_validate_get_summoner_query, convert_region_to_platform)
    def get_summoner(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> SummonerDto:
        platform_str  = query["platform"].value
        key_attr = None
        for id_type in ["name", "accountId", "id", "puuid"]:
            key_attr = query.get(id_type, None)
            if key_attr is not None:
                break
        key = "{clsname}.{platform}.{value}".format(clsname=SummonerDto.__name__,
                                                    platform=platform_str,
                                                    value=key_attr)
        return SummonerDto(self._get(key))

    @put.register(SummonerDto)
    def put_summoner(self, item: SummonerDto, context: PipelineContext = None) -> None:
        platform = Region(item["region"]).platform.value
        pairs = {}
        for id_type in ["name", "accountId", "id", "puuid"]:
            key = "{clsname}.{platform}.{value}".format(clsname=SummonerDto.__name__,
                                                        platform=platform,
                                                        value=item[id_type])
            pairs[key] = item
        self._put_many(pairs, SummonerDto)


from typing import Dict
import os

try:
    import ujson as json
except ImportError:
    import json


def load_config(filename = None) -> Dict:
    head, tail = os.path.split(__file__)

    if filename is not None:
        if not os.path.exists(filename):
            filename = os.path.normpath(os.path.join(head, filename))
            if not os.path.exists(filename):
                print("WARNING! Could not find settings file {}, using default.".format(filename))

        config = json.loads(open(filename).read())
    else:
        config = {}
    return config
